# -*- coding: utf-8 -*-
#
# File: adapters.py
#

from appy.gen import No
from appy.shared.diff import HtmlDiff
from collective.compoundcriterion.adapters import NegativePersonalLabelsAdapter
from collective.compoundcriterion.adapters import NegativePreviousIndexValuesAdapter
from collective.contact.plonegroup.utils import get_organizations
from collective.contact.plonegroup.utils import get_own_organization
from collective.contact.plonegroup.utils import get_plone_group_id
from collective.documentgenerator.adapters import GenerablePODTemplatesAdapter
from collective.eeafaceted.dashboard.adapters import DashboardGenerablePODTemplatesAdapter
from collective.eeafaceted.dashboard.content.pod_template import IDashboardPODTemplate
from collective.iconifiedcategory.adapter import CategorizedObjectAdapter
from collective.iconifiedcategory.adapter import CategorizedObjectInfoAdapter
from collective.iconifiedcategory.utils import get_categories
from datetime import datetime
from DateTime import DateTime
from eea.facetednavigation.criteria.handler import Criteria as eeaCriteria
from eea.facetednavigation.interfaces import IFacetedNavigable
from eea.facetednavigation.widgets.resultsperpage.widget import Widget as ResultsPerPageWidget
from eea.facetednavigation.widgets.storage import Criterion
from imio.actionspanel.adapters import ContentDeletableAdapter as APContentDeletableAdapter
from imio.annex.adapters import AnnexPrettyLinkAdapter
from imio.helpers import EMPTY_STRING
from imio.helpers.adapters import MissingTerms
from imio.helpers.cache import get_cachekey_volatile
from imio.helpers.cache import get_current_user_id
from imio.helpers.cache import get_plone_groups_for_user
from imio.helpers.catalog import merge_queries
from imio.helpers.content import get_user_fullname
from imio.helpers.content import get_vocab
from imio.helpers.content import get_vocab_values
from imio.helpers.content import richtextval
from imio.helpers.xhtml import xhtmlContentIsEmpty
from imio.history.adapters import BaseImioHistoryAdapter
from imio.history.adapters import ImioWfHistoryAdapter
from imio.history.interfaces import IImioHistory
from imio.history.utils import getLastAction
from imio.prettylink.adapters import PrettyLinkAdapter
from persistent.list import PersistentList
from plone import api
from plone.api.exc import InvalidParameterError
from plone.app.querystring.queryparser import parseFormquery
from plone.memoize import ram
from plone.memoize.instance import memoize
from Products.CMFCore.permissions import AccessContentsInformation
from Products.CMFCore.permissions import ModifyPortalContent
from Products.CMFCore.permissions import View
from Products.CMFCore.utils import _checkPermission
from Products.CMFPlone.utils import safe_unicode
from Products.PloneMeeting.config import AddAnnexDecision
from Products.PloneMeeting.config import ALL_ADVISERS_GROUP_VALUE
from Products.PloneMeeting.config import DUPLICATE_AND_KEEP_LINK_EVENT_ACTION
from Products.PloneMeeting.config import DUPLICATE_EVENT_ACTION
from Products.PloneMeeting.config import HIDDEN_DURING_REDACTION_ADVICE_VALUE
from Products.PloneMeeting.config import ITEM_NO_PREFERRED_MEETING_VALUE
from Products.PloneMeeting.config import NOT_GIVEN_ADVICE_VALUE
from Products.PloneMeeting.config import READER_USECASES
from Products.PloneMeeting.content.meeting import IMeeting
from Products.PloneMeeting.interfaces import IMeetingContent
from Products.PloneMeeting.MeetingConfig import CONFIGGROUPPREFIX
from Products.PloneMeeting.MeetingConfig import PROPOSINGGROUPPREFIX
from Products.PloneMeeting.MeetingConfig import READERPREFIX
from Products.PloneMeeting.MeetingConfig import SUFFIXPROFILEPREFIX
from Products.PloneMeeting.utils import compute_item_roles_to_assign_to_suffixes
from Products.PloneMeeting.utils import displaying_available_items
from Products.PloneMeeting.utils import findNewValue
from Products.PloneMeeting.utils import get_context_with_request
from Products.PloneMeeting.utils import get_dx_attrs
from Products.PloneMeeting.utils import get_referer_obj
from Products.PloneMeeting.utils import getAdvicePortalTypeIds
from Products.PloneMeeting.utils import getCurrentMeetingObject
from Products.PloneMeeting.utils import getHistoryTexts
from Products.PloneMeeting.utils import is_transition_before_date
from z3c.form.term import MissingChoiceTermsVocabulary
from zope.annotation import IAnnotations
from zope.component import getAdapter
from zope.i18n import translate
from zope.schema.vocabulary import SimpleVocabulary

import logging


logger = logging.getLogger('PloneMeeting')
CAN_NOT_DELETE_MEETING_ERROR = "This meeting can not be deleted because it contains items!"


# this catalog query will find nothing, used in CompoundCriterion adapters when necessary
def _find_nothing_query(portal_type):
    """ """
    query = {'UID': {'query': 'dummy_unexisting_uid'}, }
    return query


class AnnexContentDeletableAdapter(APContentDeletableAdapter):
    """
      Manage the mayDelete for annex and annexDecision.
      An annex/annexDecision can be deleted by users able to edit parent (item or advice).
      An annexDecision is deletable by the annexDecision Owner still able to add annexDecision.
    """
    def __init__(self, context):
        self.context = context

    def mayDelete(self, **kwargs):
        '''See docstring in interfaces.py.'''
        # check 'Delete objects' permission
        mayDelete = super(AnnexContentDeletableAdapter, self).mayDelete()
        if not mayDelete:
            parent = self.context.getParentNode()
            # able to delete an annex/annexDecision if able to edit the parent
            if _checkPermission(ModifyPortalContent, parent):
                return True

            # a 'Owner' may still remove an 'annexDecision' if enabled
            # in the cfg and if still able to add 'annexDecision'
            elif self.context.portal_type == 'annexDecision':
                tool = api.portal.get_tool('portal_plonemeeting')
                cfg = tool.getMeetingConfig(self.context)
                if cfg.getOwnerMayDeleteAnnexDecision() and \
                   _checkPermission(AddAnnexDecision, parent):
                    member = api.user.get_current()
                    if 'Owner' in member.getRolesInContext(self.context):
                        return True
        return mayDelete


class AdviceContentDeletableAdapter(APContentDeletableAdapter):
    """
      Manage the mayDelete for meetingadvice.
      Must have 'Delete objects' on the item.
      If advice was historized (advice was asked_again at least once), advice
      is not deletable.
    """
    def __init__(self, context):
        self.context = context

    def mayDelete(self, **kwargs):
        '''See docstring in interfaces.py.'''
        # check 'Delete objects' permission
        mayDelete = super(AdviceContentDeletableAdapter, self).mayDelete()
        if mayDelete:
            tool = api.portal.get_tool('portal_plonemeeting')
            if not tool.isManager(realManagers=True) and \
               getLastAction(getAdapter(self.context, IImioHistory, 'advice_given')):
                mayDelete = False
        return mayDelete


class MeetingItemContentDeletableAdapter(APContentDeletableAdapter):
    """
      Manage the mayDelete for MeetingItem.
      Must have 'Delete objects' on the item.
    """
    def __init__(self, context):
        self.context = context

    def mayDelete(self, **kwargs):
        '''See docstring in interfaces.py.'''
        # check 'Delete objects' permission
        mayDelete = super(MeetingItemContentDeletableAdapter, self).mayDelete()
        if mayDelete:
            # check itemWithGivenAdviceIsNotDeletable
            tool = api.portal.get_tool('portal_plonemeeting')
            cfg = tool.getMeetingConfig(self.context)
            if cfg.getItemWithGivenAdviceIsNotDeletable() and not tool.isManager(cfg):
                # do we have any given advice?
                # do not consider advices that are inherited
                given_advices = [advice for advice in self.context.adviceIndex.values() if
                                 not advice['inherited'] and not advice['type'] == NOT_GIVEN_ADVICE_VALUE]
                if given_advices:
                    return False
        return mayDelete


class MeetingContentDeletableAdapter(APContentDeletableAdapter):
    """
      Manage the mayDelete for Meeting.
      - must have 'Delete objects' on the meeting;
      - if user is Manager, this will remove the meeting including items;
      - if user is MeetingManager, the meeting must be empty to be removed.
    """
    def __init__(self, context):
        self.context = context

    def mayDelete(self, **kwargs):
        '''See docstring in interfaces.py.'''
        res = super(MeetingContentDeletableAdapter, self).mayDelete()
        if res:
            if self.context.number_of_items() != 0 and \
               not api.user.get_current().has_role('Manager'):
                res = No(CAN_NOT_DELETE_MEETING_ERROR)
        return res


class OrgContentDeletableAdapter(APContentDeletableAdapter):
    """
      Manage the mayDelete for organization.
      This will remove the 'Remove' button when user is on the 'own_org'.
    """
    def __init__(self, context):
        self.context = context

    def mayDelete(self, **kwargs):
        '''See docstring in interfaces.py.'''
        if not super(OrgContentDeletableAdapter, self).mayDelete():
            return False

        if self.context == get_own_organization():
            return False

        return True


class AdvicePrettyLinkAdapter(PrettyLinkAdapter):
    """ """

    def getLink_cachekey(method, self):
        '''As item title is displayed on advice, invalidate cache if item title changed.'''
        res = super(AdvicePrettyLinkAdapter, self).getLink_cachekey(self)
        # append item title
        return res + (self.context.aq_inner.aq_parent.Title(), )

    @ram.cache(getLink_cachekey)
    def getLink(self):
        """Necessary to be able to override the cachekey."""
        return self._getLink()

    def _leadingIcons(self):
        """
          Manage icons to display before the icons managed by PrettyLink._icons.
        """
        res = []
        item = self.context.aq_inner.aq_parent
        item_state = item.query_state()
        # display the waiting advices icon if relevant
        if item_state.endswith('_waiting_advices'):
            item_wf_conditions = item.wfConditions()
            if self.context.advice_group in item_wf_conditions._get_waiting_advices_icon_advisers():
                icon_name, msg = item_wf_conditions.get_waiting_advices_icon_infos()
                res.append((icon_name, msg))
        return res


class ItemPrettyLinkAdapter(PrettyLinkAdapter):
    """Override to take into account PloneMeeting use cases..."""

    def getLink_cachekey(method, self):
        '''cachekey method for self.getLink.'''
        res = super(ItemPrettyLinkAdapter, self).getLink_cachekey(self)

        # manage when displayed in availableItems on the meeting_view
        meeting_modified = None
        if displaying_available_items(self.context):
            meeting = getCurrentMeetingObject(self.context)
            if meeting:
                meeting_modified = meeting.modified()

        # manage takenOverBy
        current_member_id = None
        takenOverBy = self.context.getTakenOverBy()
        if takenOverBy:
            current_member_id = get_current_user_id(self.request)

        # manage when displaying the icon with informations about
        # the predecessor living in another MC
        predecessor_modified = None
        predecessor = self._predecessorFromOtherMC()
        if predecessor:
            predecessor_modified = predecessor.modified()

        # manage otherMC to send to, and cloned to
        # indeed we need to know where to send/have been sent if selected/unselected, ...
        ann = IAnnotations(self.context)
        other_mc_to_clone_to = [
            destMeetingConfigId for destMeetingConfigId in self.context.getOtherMeetingConfigsClonableTo()]
        destMeetingConfigIds = get_vocab_values(
            self.context, 'Products.PloneMeeting.vocabularies.other_mcs_clonable_to_vocabulary')
        other_mc_cloned_to_ann_keys = [
            destMeetingConfigId for destMeetingConfigId in destMeetingConfigIds
            if self.context._getSentToOtherMCAnnotationKey(destMeetingConfigId) in ann]

        # an advice WF state changed, this is useful for the waiting_advices icon
        # changing advice review_state will change advice._p_mtime
        item_state = res[3]
        advice_modified = None
        if item_state.endswith('_waiting_advices'):
            advices = self.context.getAdvices()
            if advices:
                advice_modified = max([advice._p_mtime for advice in advices])

        return res + (meeting_modified,
                      advice_modified,
                      takenOverBy,
                      current_member_id,
                      predecessor_modified,
                      other_mc_to_clone_to,
                      other_mc_cloned_to_ann_keys)

    @ram.cache(getLink_cachekey)
    def getLink(self):
        """Necessary to be able to override the cachekey."""
        return self._getLink()

    def _predecessorFromOtherMC(self):
        predecessor = self.context.get_predecessor()
        if predecessor and predecessor.portal_type != self.context.portal_type:
            return predecessor
        return None

    def _leadingIcons(self):
        """
          Manage icons to display before the icons managed by PrettyLink._icons.
        """
        res = []

        self.tool = api.portal.get_tool('portal_plonemeeting')
        self.cfg = self.tool.getMeetingConfig(self.context)
        usedItemAttributes = self.cfg.getUsedItemAttributes()
        usedMeetingAttributes = self.cfg.getUsedMeetingAttributes()

        if displaying_available_items(self.context):
            meeting = getCurrentMeetingObject(self.context)
            # there could be no meeting if we opened an item from the available items view
            if meeting:
                # Item is in the list of available items, check if we
                # must show a deadline- or late-related icon.
                if self.context.wfConditions().isLateFor(meeting):
                    # A late item, or worse: a late item not respecting the freeze deadline.
                    if "freeze_deadline" in usedMeetingAttributes and \
                       getattr(meeting, "freeze_deadline", None) is not None and \
                       not is_transition_before_date(
                            self.context, "validate", meeting.freeze_deadline):
                        res.append(('deadlineKo.png',
                                    translate('icon_help_freeze_deadline_ko',
                                              domain="PloneMeeting",
                                              context=self.request)))
                    else:
                        res.append(('late.png',
                                    translate('icon_help_late',
                                              domain="PloneMeeting",
                                              context=self.request)))
                elif meeting.query_state() == 'created' and \
                        "validation_deadline" in usedMeetingAttributes and \
                        getattr(meeting, "validation_deadline", None) is not None and \
                        not is_transition_before_date(
                            self.context, "validate", meeting.validation_deadline):
                    res.append(('deadlineKo.png',
                                translate('icon_help_validation_deadline_ko',
                                          domain="PloneMeeting",
                                          context=self.request)))

        self.itemState = self.context.query_state()
        # specifically manage states without leading icons to speed up things
        if self.itemState in ('itemcreated', 'proposed', 'validated'):
            pass
        elif self.itemState == 'delayed':
            res.append(('delayed.png', translate('icon_help_delayed',
                                                 domain="PloneMeeting",
                                                 context=self.request)))
        elif self.itemState == 'refused':
            res.append(('refused.png', translate('icon_help_refused',
                                                 domain="PloneMeeting",
                                                 context=self.request)))
        elif self.itemState == 'returned_to_proposing_group':
            res.append(('return_to_proposing_group.png',
                        translate('icon_help_returned_to_proposing_group',
                                  domain="PloneMeeting",
                                  context=self.request)))
        elif self.itemState == 'accepted_but_modified':
            res.append(('accepted_but_modified.png', translate('icon_help_accepted_but_modified',
                                                               domain="PloneMeeting",
                                                               context=self.request)))
        elif self.itemState == 'accepted_out_of_meeting':
            res.append(('accept_out_of_meeting.png',
                        translate('icon_help_accepted_out_of_meeting',
                                  domain="PloneMeeting",
                                  context=self.request,
                                  default=translate('accepted_out_of_meeting',
                                                    domain="plone",
                                                    context=self.request))))
        elif self.itemState == 'accepted_out_of_meeting_emergency':
            res.append(('accept_out_of_meeting_emergency.png',
                        translate('icon_help_accepted_out_of_meeting_emergency',
                                  domain="PloneMeeting",
                                  context=self.request,
                                  default=translate('accepted_out_of_meeting_emergency',
                                                    domain="plone",
                                                    context=self.request))))
        elif self.itemState == 'transfered':
            res.append(('transfer.png',
                        translate('icon_help_transfered',
                                  domain="PloneMeeting",
                                  context=self.request,
                                  default=translate('transfered',
                                                    domain="plone",
                                                    context=self.request))))
        elif self.itemState == 'pre_accepted':
            res.append(('pre_accepted.png', translate('icon_help_pre_accepted',
                                                      domain="PloneMeeting",
                                                      context=self.request)))
        elif self.itemState == 'postponed_next_meeting':
            res.append(('postponed_next_meeting.png',
                        translate('icon_help_postponed_next_meeting',
                                  domain="PloneMeeting",
                                  context=self.request)))
        elif self.itemState == 'marked_not_applicable':
            res.append(('marked_not_applicable.png',
                        translate('icon_help_marked_not_applicable',
                                  domain="PloneMeeting",
                                  context=self.request)))
        elif self.itemState == 'removed':
            res.append(('removed.png',
                        translate('icon_help_removed',
                                  domain="PloneMeeting",
                                  context=self.request)))
        elif self.itemState.endswith('_waiting_advices'):
            icon_name, msg = self.context.wfConditions().get_waiting_advices_icon_infos()
            res.append((icon_name, msg))
        elif self.itemState.startswith('returned_to_proposing_group_'):
            # get info about return_to_proposing_group validation
            # level in MeetingConfig.itemWFValidationLevels
            validation_state = self.itemState.replace('returned_to_proposing_group_', '')
            level = self.cfg.getItemWFValidationLevels(states=[validation_state])
            res.append(
                ('goTo_{0}.png'.format(self.itemState),
                 translate('returned_to_proposing_group_with_validation_state',
                           domain="plone",
                           mapping={"validation_state":
                                    translate(
                                        safe_unicode(level['state_title']),
                                        domain='plone',
                                        context=self.request), },
                           context=self.request)))
        else:
            # manage MeetingConfig.itemWFValidationLevels states
            item_validation_states = self.cfg.getItemWFValidationLevels(data='state', only_enabled=True)
            if self.itemState in item_validation_states:
                level = self.cfg.getItemWFValidationLevels(states=[self.itemState], only_enabled=True)
                res.append(('{0}.png'.format(level['leading_transition']),
                            translate('icon_help_{0}'.format(self.itemState),
                                      domain="PloneMeeting",
                                      context=self.request,
                                      default=translate(safe_unicode(level['state_title']),
                                                        domain='plone',
                                                        context=self.request))))

        # Display icons about sent/cloned to other meetingConfigs
        clonedToOtherMCIds = self.context._getOtherMeetingConfigsImAmClonedIn()
        toLocalizedTime = None
        for clonedToOtherMCId in clonedToOtherMCIds:
            # Append a tuple with name of the icon and a list containing
            # the msgid and the mapping as a dict
            # if item sent to the other mc is inserted into a meeting,
            # we display the meeting date
            emergency = clonedToOtherMCId in self.context.getOtherMeetingConfigsClonableToEmergency()
            clonedToOtherMC = self.tool.get(clonedToOtherMCId)
            msgid = emergency and 'sentto_othermeetingconfig_emergency' or 'sentto_othermeetingconfig'
            msg = translate(
                msgid,
                mapping={
                    'meetingConfigTitle':
                        safe_unicode(clonedToOtherMC.Title(include_config_group=True))},
                domain="PloneMeeting",
                context=self.request)

            clonedBrain = self.context.getItemClonedToOtherMC(clonedToOtherMCId, theObject=False)
            # do not check on meeting_date because it may contains '1950/01/01',
            # see meeting_date indexer in indexes.py
            if clonedBrain.meeting_uid != ITEM_NO_PREFERRED_MEETING_VALUE:
                # avoid instantiating toLocalizedTime more than once
                toLocalizedTime = toLocalizedTime or self.context.restrictedTraverse('@@plone').toLocalizedTime
                long_format = clonedBrain.meeting_date.hour and True or False
                msg = msg + u' ({0})'.format(toLocalizedTime(
                    clonedBrain.meeting_date, long_format=long_format))
            iconName = emergency and "clone_to_other_mc_emergency" or "clone_to_other_mc"
            # manage the otherMeetingConfigsClonableToPrivacy
            if 'privacy' in clonedToOtherMC.getUsedItemAttributes():
                iconName += "_{0}".format(clonedBrain.privacy)
                msg = msg + u' ({0})'.format(translate(clonedBrain.privacy,
                                                       domain="PloneMeeting",
                                                       context=self.request))
            res.append(("{0}.png".format(iconName), msg))

        # if not already cloned to another mc, maybe it will be?
        # we could have an item to clone to 2 other MCs, one already sent, not the other...
        otherMeetingConfigsClonableTo = self.context.getOtherMeetingConfigsClonableTo()
        for otherMeetingConfigClonableToId in otherMeetingConfigsClonableTo:
            # already cloned?
            if otherMeetingConfigClonableToId in clonedToOtherMCIds:
                continue

            # Append a tuple with name of the icon and a list containing
            # the msgid and the mapping as a dict
            otherMeetingConfigClonableTo = self.tool.get(otherMeetingConfigClonableToId)
            emergency = otherMeetingConfigClonableToId in self.context.getOtherMeetingConfigsClonableToEmergency()
            msgid = emergency and 'will_be_sentto_othermeetingconfig_emergency' or \
                'will_be_sentto_othermeetingconfig'
            iconName = emergency and "will_be_cloned_to_other_mc_emergency" or "will_be_cloned_to_other_mc"
            msg = translate(msgid,
                            mapping={'meetingConfigTitle': safe_unicode(
                                     otherMeetingConfigClonableTo.Title(include_config_group=True))},
                            domain="PloneMeeting",
                            context=self.request)
            # manage the otherMeetingConfigsClonableToPrivacy
            suffix = ''
            if 'otherMeetingConfigsClonableToPrivacy' in usedItemAttributes and \
               'privacy' in otherMeetingConfigClonableTo.getUsedItemAttributes():
                if otherMeetingConfigClonableToId in self.context.getOtherMeetingConfigsClonableToPrivacy():
                    suffix = "_secret"
                else:
                    suffix = "_public"
                msg = msg + u' ({0})'.format(translate(suffix[1:],
                                                       domain="PloneMeeting",
                                                       context=self.request))
            res.append(("{0}{1}.png".format(iconName, suffix),
                        msg))

        # display an icon if item is sent from another mc
        predecessor = self._predecessorFromOtherMC()
        if predecessor:
            predecessorCfg = self.tool.getMeetingConfig(predecessor)
            predecessorMeeting = predecessor.getMeeting()
            predecessor_state = predecessor.query_state()
            translated_state = translate(predecessor_state, domain='plone', context=self.request)
            if not predecessorMeeting:
                res.append(
                    ('cloned_not_decided.png',
                     translate(
                        'icon_help_cloned_not_presented',
                        domain="PloneMeeting",
                        mapping={
                            'meetingConfigTitle':
                            safe_unicode(predecessorCfg.Title(include_config_group=True)),
                            'predecessorState': translated_state},
                        context=self.request,
                        default="Sent from ${meetingConfigTitle}, "
                        "original item is \"${predecessorState}\".")))
            else:
                if predecessor_state in predecessorCfg.getItemPositiveDecidedStates():
                    res.append(
                        ('cloned_and_decided.png',
                         translate(
                            'icon_help_cloned_and_decided',
                            mapping={
                                'meetingDate': self.tool.format_date(predecessorMeeting.date),
                                'meetingConfigTitle':
                                safe_unicode(predecessorCfg.Title(include_config_group=True)),
                                'predecessorState': translated_state},
                            domain="PloneMeeting",
                            context=self.request,
                            default="Sent from ${meetingConfigTitle} (${meetingDate}), "
                            "original item is \"${predecessorState}\".")))
                else:
                    res.append(
                        ('cloned_not_decided.png',
                         translate(
                            'icon_help_cloned_not_decided',
                            mapping={
                                'meetingDate': self.tool.format_date(predecessorMeeting.date),
                                'meetingConfigTitle':
                                safe_unicode(predecessorCfg.Title(include_config_group=True)),
                                'predecessorState': translated_state},
                            domain="PloneMeeting",
                            context=self.request,
                            default="Sent from ${meetingConfigTitle} (${meetingDate}), "
                            "original item is \"${predecessorState}\".")))

        # display icons if element is down the workflow or up for at least second time...
        # display it only for items before state 'validated'
        downOrUpWorkflowAgain = self.context.downOrUpWorkflowAgain()
        if downOrUpWorkflowAgain == "down":
            res.append(('wf_down.png', translate('icon_help_wf_down',
                                                 domain="PloneMeeting",
                                                 context=self.request)))
        elif downOrUpWorkflowAgain == "up":
            res.append(('wf_up.png', translate('icon_help_wf_up',
                                               domain="PloneMeeting",
                                               context=self.request)))

        # In some cases, it does not matter if an item is inMeeting or not.
        if 'oralQuestion' in usedItemAttributes:
            if self.context.getOralQuestion():
                res.append(('oralQuestion.png', translate('this_item_is_an_oral_question',
                                                          domain="PloneMeeting",
                                                          context=self.request)))
        if 'emergency' in usedItemAttributes:
            # display an icon if emergency asked/accepted/refused
            itemEmergency = self.context.getEmergency()
            if itemEmergency == 'emergency_asked':
                res.append(('emergency_asked.png', translate('emergency_asked',
                                                             domain="PloneMeeting",
                                                             context=self.request)))
            elif itemEmergency == 'emergency_accepted':
                res.append(('emergency_accepted.png', translate('emergency_accepted',
                                                                domain="PloneMeeting",
                                                                context=self.request)))
            elif itemEmergency == 'emergency_refused':
                res.append(('emergency_refused.png', translate('emergency_refused',
                                                               domain="PloneMeeting",
                                                               context=self.request)))
        if 'takenOverBy' in usedItemAttributes:
            takenOverBy = self.context.getTakenOverBy()
            if takenOverBy:
                # if taken over, display a different icon if taken over by current user or not
                user_id = get_current_user_id(self.request)
                takenOverByCurrentUser = bool(user_id == takenOverBy)
                iconName = takenOverByCurrentUser and 'takenOverByCurrentUser.png' or 'takenOverByOtherUser.png'
                res.append((iconName, translate(u'Taken over by ${fullname}',
                                                domain="PloneMeeting",
                                                mapping={'fullname': get_user_fullname(takenOverBy)},
                                                context=self.request)))

        if self.context.getIsAcceptableOutOfMeeting():
            res.append(('acceptable_out_of_meeting.png',
                        translate('icon_help_isAcceptableOutOfMeeting',
                                  domain="PloneMeeting",
                                  context=self.request)))
        return res


class MeetingPrettyLinkAdapter(PrettyLinkAdapter):
    """
      Override to take into account PloneMeeting use cases...
    """
    def getLink_cachekey(method, self):
        '''cachekey method for self.getLink.'''
        res = super(MeetingPrettyLinkAdapter, self).getLink_cachekey(self)

        # res check context_modified but we just want to check date
        meeting_date = self.context.date
        res = list(res)
        del res[1]
        res.append(meeting_date)
        # check also on usecases adding an icon,
        res.append(self.context.extraordinary_session)
        res.append(self.context.videoconference)
        res.append(self.context.adopts_next_agenda_of)
        return tuple(res)

    @ram.cache(getLink_cachekey)
    def getLink(self):
        """Necessary to be able to override the cachekey."""
        return self._getLink()

    def _leadingIcons(self):
        """
          Manage icons to display before the icons managed by PrettyLink._icons.
        """
        res = []
        if self.context.extraordinary_session:
            res.append(('extraordinarySession.png',
                        translate('this_meeting_is_extraordinary_session',
                                  domain="PloneMeeting",
                                  context=self.request)))
        if self.context.videoconference:
            res.append(('videoconference.png',
                        translate('this_meeting_is_videoconference',
                                  domain="PloneMeeting",
                                  context=self.request)))
        if self.context.adopts_next_agenda_of:
            tool = api.portal.get_tool('portal_plonemeeting')
            res.append(
                ('adopts_next_agenda_of.png',
                 translate(
                    'this_meeting_adopts_next_agenda_of',
                    mapping={
                        'cfg_titles': u", ".join([
                            safe_unicode(tool.get(cfg_id).Title(include_config_group=True))
                            for cfg_id in self.context.adopts_next_agenda_of])},
                    domain="PloneMeeting",
                    context=self.request)))
        return res


class PMAnnexPrettyLinkAdapter(AnnexPrettyLinkAdapter):
    """
      Override to take into account PloneMeeting use cases...
    """

    def getLink_cachekey(method, self):
        '''cachekey method for self.getLink.'''
        res = super(PMAnnexPrettyLinkAdapter, self).getLink_cachekey(self)
        # check also MeetingConfig modified as updating categorized elements
        # from the ContentCategory will update MeetingConfig.modified
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(self.context.aq_parent)
        return res + (cfg.modified(), )

    @ram.cache(getLink_cachekey)
    def getLink(self):
        """Necessary to be able to override the cachekey."""
        return self._getLink()


class PMWfHistoryAdapter(ImioWfHistoryAdapter):
    """
      Override the imio.history ImioHistoryAdapter.
    """

    def __init__(self, context):
        super(PMWfHistoryAdapter, self).__init__(context)
        self.tool = api.portal.get_tool('portal_plonemeeting')
        self.cfg = self.tool.getMeetingConfig(self.context)

    def ignorableHistoryComments(self):
        """Add some more ignorable history comments."""
        ignorable_history_comment = super(PMWfHistoryAdapter, self).ignorableHistoryComments()
        ignorable_history_comment += (u'create_meeting_item_from_template_comments',
                                      u'create_from_predecessor_comments',
                                      u'{0}_comments'.format(DUPLICATE_AND_KEEP_LINK_EVENT_ACTION),
                                      u'{0}_comments'.format(DUPLICATE_EVENT_ACTION),
                                      u'wf_transition_triggered_by_application')
        return ignorable_history_comment

    def mayViewComment(self, event):
        """
          By default, every p_event comment is viewable except for MeetingItem, if
          'hideItemHistoryCommentsToUsersOutsideProposingGroup' is enabled in the MeetingConfig,
          only members of the group managing item at event['review_state'] will be able to access
          history comment.
        """
        userMayAccessComment = True
        if self.context.getTagName() == 'MeetingItem':
            if self.cfg.getHideItemHistoryCommentsToUsersOutsideProposingGroup() and \
               not self.tool.isManager(self.cfg):
                userOrgUids = self.tool.get_orgs_for_user()
                group_managing_item_uid = \
                    self.context.adapted()._getGroupManagingItem(event['review_state'])
                if group_managing_item_uid not in userOrgUids:
                    userMayAccessComment = False
        return userMayAccessComment

    def get_history_data(self):
        """WF history is mixed with data_changes history."""
        history = super(PMWfHistoryAdapter, self).get_history_data()
        res = []
        for event in history:
            new_event = event.copy()
            if new_event['action'] != '_datachange_':
                res.append(new_event)
        return res


class PMDataChangesHistoryAdapter(ImioWfHistoryAdapter):
    """ """

    history_type = 'data_changes'
    highlight_last_comment = False

    def get_history_data(self):
        """WF history is mixed with data_changes history."""
        history = super(PMDataChangesHistoryAdapter, self).get_history_data()
        full_datachanges_history = []
        # first pass, keep datachanges
        for event in history:
            new_event = event.copy()
            if new_event['action'] == '_datachange_':
                full_datachanges_history.append(new_event)

        # second pass, compute datachanges
        res = []
        i = -1
        full_datachanges_history.reverse()
        while (i + 1) < len(full_datachanges_history):
            i += 1
            new_event = full_datachanges_history[i].copy()
            new_event['changes'] = {}
            new_event['type'] = self.history_type
            for name, oldValue in full_datachanges_history[i]['changes'].iteritems():
                widgetName = self.context.getField(name).widget.getName()
                if widgetName == 'RichWidget':
                    if xhtmlContentIsEmpty(oldValue):
                        val = '-'
                    else:
                        newValue = findNewValue(self.context, name, full_datachanges_history, i - 1)
                        # Compute the diff between oldValue and newValue
                        iMsg, dMsg = getHistoryTexts(self.context, new_event)
                        comparator = HtmlDiff(oldValue, newValue, iMsg, dMsg)
                        val = comparator.get()
                    new_event['changes'][name] = val
                elif widgetName == 'BooleanWidget':
                    label = oldValue and 'Yes' or 'No'
                    new_event['changes'][name] = translate(label, domain="plone", context=self.request)
                elif widgetName == 'TextAreaWidget':
                    val = oldValue.replace('\r', '').replace('\n', '<br/>')
                    new_event['changes'][name] = val
                elif widgetName == 'SelectionWidget':
                    allValues = self.context.getField(name).Vocabulary(self.context)
                    val = allValues.getValue(oldValue or '')
                    new_event['changes'][name] = val or '-'
                elif widgetName == 'MultiSelectionWidget':
                    allValues = self.context.getField(name).Vocabulary(self.context)
                    val = [allValues.getValue(v) for v in oldValue]
                    # remove None in val in case we have old values that
                    # does not exist anymore in allValues
                    val = [v for v in val if v is not None]
                    if not val:
                        val = '-'
                    else:
                        val = '<br/>'.join(val)
                    new_event['changes'][name] = val
                else:
                    new_event['changes'][name] = oldValue
            res.append(new_event)
        return res


class PMEmergencyChangesHistoryAdapter(BaseImioHistoryAdapter):
    """ """

    history_type = 'emergency_changes'
    history_attr_name = 'emergency_changes_history'


class PMCompletenessChangesHistoryAdapter(BaseImioHistoryAdapter):
    """ """

    history_type = 'completeness_changes'
    history_attr_name = 'completeness_changes_history'


class PMAdviceHideDuringRedactionHistoryAdapter(BaseImioHistoryAdapter):
    """ """

    history_type = 'advice_hide_during_redaction'
    history_attr_name = 'advice_hide_during_redaction_history'


class PMAdviceGivenHistoryAdapter(BaseImioHistoryAdapter):
    """ """

    history_type = 'advice_given'
    history_attr_name = 'advice_given_history'

    def revert_to_last_event(self):
        """Revert advice values to last historized event."""
        last_action = getLastAction(self)
        rich_text_field_names = get_dx_attrs(
            self.context.portal_type, richtext_only=True, as_display_list=False)
        for field_data in last_action['advice_data']:
            # handle rich text to store a RichTextValue
            field_value = field_data['field_value']
            if field_data['field_name'] in rich_text_field_names:
                field_value = richtextval(field_value)
            setattr(self.context, field_data['field_name'], field_value)


class Criteria(eeaCriteria):
    """
      Override method that gets criteria to be able to manage various use cases :
      - for meetings : get the criteria from the MeetingConfig (searches_items) and filter
        out elements not in MeetingConfig.getDashboardAvailableItemsFilters and not in
        MeetingConfig.getDashboardPresentedItemsFilters;
      - for listing of items : filter out criteria no in MeetingConfig.getDashboardItemsFilters;
      - for listing of meetings : filter out criteria no in MeetingConfig.getDashboardMeetingsFilters.
    """

    def __init__(self, context):
        """ """
        self.context, self.criteria = self.compute_criteria(context)

    def compute_criteria(self, context):
        """ """
        req = context.REQUEST
        key = "PloneMeeting-adapters-criteria-compute_criteria-{0}".format(repr(context))
        cache = IAnnotations(req)
        cached = cache.get(key, None)

        if cached is not None:
            self.context, self.criteria = cached
        else:
            # return really stored widgets when necessary
            if 'portal_plonemeeting' in context.absolute_url() or \
               req.get('enablingFacetedDashboard', False) or \
               (req.get('SERVER_URL') == 'http://foo' or
                    req.get('PARENTS', [])[0] == api.portal.get_tool('portal_setup')):  # migrating
                super(Criteria, self).__init__(context)
                return self.context, self.criteria
            try:
                tool = api.portal.get_tool('portal_plonemeeting')
            except InvalidParameterError:
                # in case 'portal_plonemeeting' is not available, use original criteria behaviour
                super(Criteria, self).__init__(context)
                return self.context, self.criteria
            cfg = tool.getMeetingConfig(context)
            if not cfg:
                super(Criteria, self).__init__(context)
                return self.context, self.criteria
            else:
                # meeting view
                kept_filters = []
                resultsperpagedefault = 20
                meeting_view = False
                compute = True
                if IMeeting.providedBy(context):
                    meeting_view = True
                    is_displaying_available_items = displaying_available_items(context)
                    self.context = cfg.searches.searches_items
                    if is_displaying_available_items:
                        kept_filters = cfg.getDashboardMeetingAvailableItemsFilters()
                        resultsperpagedefault = cfg.getMaxShownAvailableItems()
                    else:
                        kept_filters = cfg.getDashboardMeetingLinkedItemsFilters()
                        resultsperpagedefault = cfg.getMaxShownMeetingItems()
                else:
                    # on a faceted?  it is a pmFolder or a subFolder of the pmFolder
                    resultsperpagedefault = cfg.getMaxShownListings()
                    if IFacetedNavigable.providedBy(context):
                        # keep relevant filters depending on configuration
                        if context.getId() == 'searches_items':
                            kept_filters = cfg.getDashboardItemsListingsFilters()
                            self.context = cfg.searches.searches_items
                        elif context.getId() == 'searches_meetings':
                            kept_filters = cfg.getDashboardMeetingsListingsFilters()
                            self.context = cfg.searches.searches_meetings
                        elif context.getId() == 'searches_decisions':
                            kept_filters = cfg.getDashboardMeetingsListingsFilters()
                            self.context = cfg.searches.searches_decisions
                        else:
                            compute = False
                            self.context = cfg.searches
                            self.criteria = self._criteria()

                if compute:
                    res = PersistentList()
                    for criterion in self._criteria():
                        # take care that we have the stored criteria here so
                        # if one need to be changed, we must create a
                        # new Criterion or the stored value is changed!
                        if meeting_view and criterion.widget == u'sorting':
                            # keep it only of displaying available items, default sorting
                            # is set on 'getProposingGroup', if not displaying available items
                            # the sorting widget is not kept so sorting is disabled for presented items
                            if is_displaying_available_items:
                                new_criterion = Criterion()
                                new_criterion.update(**criterion.__dict__)
                                new_criterion.default = u'getProposingGroup'
                                res.append(new_criterion)
                            continue
                        # ignore the collection widget when on meeting_view
                        if meeting_view and criterion.widget == u'collection-link':
                            new_criterion = Criterion()
                            new_criterion.update(**criterion.__dict__)
                            new_criterion.default = u''
                            res.append(new_criterion)
                            continue

                        if criterion.section != u'advanced' or \
                           criterion.__name__ in kept_filters:
                            # create new object to avoid modifying stored one
                            new_criterion = Criterion()
                            new_criterion.update(**criterion.__dict__)
                            # manage default value for the 'resultsperpage' criterion
                            if criterion.widget == ResultsPerPageWidget.widget_type:
                                new_criterion.default = resultsperpagedefault
                            res.append(new_criterion)
                            continue
                    self.criteria = res
            cache[key] = self.context, self.criteria
        return self.context, self.criteria


class CompoundCriterionBaseAdapter(object):

    def __init__(self, context):
        self.context = context
        self.request = self.context.REQUEST
        self.tool = api.portal.get_tool('portal_plonemeeting')
        self.cfg = self.tool.getMeetingConfig(self.context)


# cachekeys useable for CompoundCriterionAdapters
def query_user_groups_cachekey(method, self):
    '''cachekey method for caching as long as global users/groups
       associations did not change.'''
    # always check cfg.modified() as queries are portal_type aware
    cfg_modified = self.cfg and self.cfg.modified() or datetime.now()
    return self.context.modified(), get_current_user_id(self.request), \
        get_cachekey_volatile('_users_groups_value'), cfg_modified


def query_meeting_config_modified_cachekey(method, self):
    '''cachekey method for caching as long as MeetingConfig not modified.'''
    cfg_modified = self.cfg and self.cfg.modified() or datetime.now()
    return self.context.modified(), cfg_modified


def forever_cachekey(method, self):
    '''cachekey method for caching forever until cache is invalidated.'''
    return True


class LastDecisionsAdapter(CompoundCriterionBaseAdapter):

    @property
    def query_last_decisions(self):
        '''Patch the query 'meeting_date' to not limit the search
           to 'today' but 60 days in the future.'''
        if not self.cfg:
            return {}
        # remove the 'last-decisions' adapter from query so we may parse it
        # or it will lead to a RuntimeError: maximum recursion depth exceeded
        query = [term for term in self.context.query if term[u'i'] != u'CompoundCriterion']
        parsedQuery = parseFormquery(self.context, query)
        # change the second date of meeting_date query, aka the 'max' date
        # use DateTime because 'plone.app.querystring.operation.date.largerThanRelativeDate'
        # will use a DateTime
        parsedQuery['meeting_date']['query'][1] = DateTime() + 60
        return parsedQuery

    # we may not ram.cache methods in same file with same name...
    query = query_last_decisions


class ItemsOfMyGroupsAdapter(CompoundCriterionBaseAdapter):

    @property
    @ram.cache(query_user_groups_cachekey)
    def query_itemsofmygroups(self):
        '''Queries all items of groups of the current user, no matter wich suffix
           of the group the user is in.'''
        if not self.cfg:
            return {}
        userOrgUids = self.tool.get_orgs_for_user(only_selected=False)
        return {'portal_type': {'query': self.cfg.getItemTypeName()},
                'getProposingGroup': {'query': userOrgUids}, }

    # we may not ram.cache methods in same file with same name...
    query = query_itemsofmygroups


class MyItemsTakenOverAdapter(CompoundCriterionBaseAdapter):

    @property
    # @ram.cache(forever_cachekey)
    def query_myitemstakenover(self):
        '''Queries all items that current user take over.'''
        if not self.cfg:
            return {}
        member_id = get_current_user_id(self.request)
        return {'portal_type': {'query': self.cfg.getItemTypeName()},
                'getTakenOverBy': {'query': member_id}, }

    # we may not ram.cache methods in same file with same name...
    query = query_myitemstakenover


class ItemsInCopyAdapter(CompoundCriterionBaseAdapter):

    @property
    @ram.cache(query_user_groups_cachekey)
    def query_itemsincopy(self):
        '''Queries all items for which the current user is in copyGroups.'''
        if not self.cfg:
            return {}
        userPloneGroups = get_plone_groups_for_user()
        return {'portal_type': {'query': self.cfg.getItemTypeName()},
                # KeywordIndex 'getCopyGroups' use 'OR' by default
                'getCopyGroups': {'query': userPloneGroups}, }

    # we may not ram.cache methods in same file with same name...
    query = query_itemsincopy


class BaseItemsToValidateOfHighestHierarchicLevelAdapter(CompoundCriterionBaseAdapter):

    def _query(self, prefix_review_state=''):
        '''Return a list of items that the user can validate regarding his highest hierarchic level.
           So if a user is 'prereviewer' and 'reviewier', the search will only return items
           in state corresponding to his 'reviewer' role.'''
        if not self.cfg:
            return {}

        # now get highest hierarchic level for every user groups
        org_uids = self.tool.get_orgs_for_user()
        userPloneGroupIds = get_plone_groups_for_user()
        reviewers = self.cfg.reviewersFor()
        userReviewerPloneGroupIds = []
        for org_uid in org_uids:
            for reviewer_level in reviewers:
                plone_group_id = get_plone_group_id(org_uid, reviewer_level)
                if plone_group_id in userPloneGroupIds:
                    userReviewerPloneGroupIds.append(plone_group_id)
                    break

        reviewProcessInfos = []
        for plone_group_id in userReviewerPloneGroupIds:
            # append group name without suffix
            org_uid, reviewer_level = plone_group_id.split('_')
            # reviewers[reviewer_level] is a list of states
            reviewProcessInfo = [
                '{0}__reviewprocess__{1}'.format(
                    org_uid,
                    '{0}{1}'.format(prefix_review_state, review_state))
                for review_state in reviewers[reviewer_level]
            ]
            reviewProcessInfos.extend(reviewProcessInfo)
        return {'portal_type': {'query': self.cfg.getItemTypeName()},
                'reviewProcessInfo': {'query': reviewProcessInfos}, }


class ItemsToValidateOfHighestHierarchicLevelAdapter(
        BaseItemsToValidateOfHighestHierarchicLevelAdapter):

    @property
    @ram.cache(query_user_groups_cachekey)
    def query_itemstovalidateofhighesthierarchiclevel(self):
        return self._query()

    # we may not ram.cache methods in same file with same name...
    query = query_itemstovalidateofhighesthierarchiclevel


class ItemsToCorrectToValidateOfHighestHierarchicLevelAdapter(
        BaseItemsToValidateOfHighestHierarchicLevelAdapter):

    @property
    @ram.cache(query_user_groups_cachekey)
    def query_itemstocorrecttovalidateofhighesthierarchiclevel(self):
        return self._query(prefix_review_state='returned_to_proposing_group_')

    # we may not ram.cache methods in same file with same name...
    query = query_itemstocorrecttovalidateofhighesthierarchiclevel


class AllItemsToValidateOfHighestHierarchicLevelAdapter(
        BaseItemsToValidateOfHighestHierarchicLevelAdapter):

    @property
    @ram.cache(query_user_groups_cachekey)
    def query_allitemstovalidateofhighesthierarchiclevel(self):
        to_validate_query = self._query()
        to_correct_to_validate_query = self._query(prefix_review_state='returned_to_proposing_group_')
        query = merge_queries([to_validate_query, to_correct_to_validate_query])
        return query

    # we may not ram.cache methods in same file with same name...
    query = query_allitemstovalidateofhighesthierarchiclevel


class BaseItemsToValidateOfEveryReviewerLevelsAndLowerLevelsAdapter(CompoundCriterionBaseAdapter):

    def _query(self, prefix_review_state=''):
        '''This will check for user highest reviewer level of each of his groups and return these items and
           items of lower reviewer levels.
           This search works if the workflow manage reviewer levels where higher reviewer level
           can validate lower reviewer levels EVEN IF THE USER IS NOT IN THE CORRESPONDING PLONE SUBGROUP.
           For example with a 3 levels reviewer workflow, called review1 (lowest level),
           review2 and review3 (highest level) :
           - reviewer1 may validate items in reviewer1;
           - reviewer2 may validate items in reviewer1 and reviewer2;
           - reviewer3 may validate items in reviewer1, reviewer2 and reviewer3.
           So get highest hierarchic level of each group of the user and
           take into account lowest levels too.'''
        if not self.cfg:
            return {}
        # search every highest reviewer level for each group of the user
        user_org_uids = self.tool.get_orgs_for_user()
        reviewProcessInfos = set([])
        for org_uid in user_org_uids:
            ploneGroups = self.tool.get_filtered_plone_groups_for_user(org_uids=[org_uid])
            # now that we have Plone groups of the organization
            # we can get highest hierarchic level and find sub levels
            highestReviewerLevel = self.cfg._highestReviewerLevel(ploneGroups)
            if not highestReviewerLevel:
                continue
            foundLevel = False
            reviewers = self.cfg.reviewersFor()
            for reviewer_suffix, review_states in reviewers.items():
                if not foundLevel and not reviewer_suffix == highestReviewerLevel:
                    continue
                foundLevel = True
                review_states = [
                    '%s%s' % (prefix_review_state, review_state) for review_state in review_states]
                reviewProcessInfo = [
                    '%s__reviewprocess__%s' % (org_uid, review_state) for review_state in review_states]
                reviewProcessInfos.update(set(reviewProcessInfo))
        if not reviewProcessInfos:
            # in this case, we do not want to display a result
            # we return an unknown review_state
            return _find_nothing_query(self.cfg.getItemTypeName())
        return {'portal_type': {'query': self.cfg.getItemTypeName()},
                'reviewProcessInfo': {'query': sorted(reviewProcessInfos)}, }


class ItemsToValidateOfEveryReviewerLevelsAndLowerLevelsAdapter(
        BaseItemsToValidateOfEveryReviewerLevelsAndLowerLevelsAdapter):

    @property
    @ram.cache(query_user_groups_cachekey)
    def query_itemstovalidateofeveryreviewerlevelsandlowerlevels(self):
        return self._query()

    # we may not ram.cache methods in same file with same name...
    query = query_itemstovalidateofeveryreviewerlevelsandlowerlevels


class ItemsToCorrectToValidateOfEveryReviewerLevelsAndLowerLevelsAdapter(
        BaseItemsToValidateOfEveryReviewerLevelsAndLowerLevelsAdapter):

    @property
    @ram.cache(query_user_groups_cachekey)
    def query_itemstocorrecttovalidateofeveryreviewerlevelsandlowerlevels(self):
        return self._query(prefix_review_state='returned_to_proposing_group_')

    # we may not ram.cache methods in same file with same name...
    query = query_itemstocorrecttovalidateofeveryreviewerlevelsandlowerlevels


class AllItemsToValidateOfEveryReviewerLevelsAndLowerLevelsAdapter(
        BaseItemsToValidateOfEveryReviewerLevelsAndLowerLevelsAdapter):

    @property
    @ram.cache(query_user_groups_cachekey)
    def query_allitemstovalidateofeveryreviewerlevelsandlowerlevels(self):
        to_validate_query = self._query()
        to_correct_to_validate_query = self._query(prefix_review_state='returned_to_proposing_group_')
        query = merge_queries([to_validate_query, to_correct_to_validate_query])
        return query

    # we may not ram.cache methods in same file with same name...
    query = query_allitemstovalidateofeveryreviewerlevelsandlowerlevels


class BaseItemsToValidateOfMyReviewerGroupsAdapter(CompoundCriterionBaseAdapter):

    def _query(self, prefix_review_state=''):
        '''Return a list of items that the user could validate.  So it returns every items the current
           user is able to validate at any state of the validation process.  So if a user is 'prereviewer'
           and 'reviewer' for a group, the search will return items in both states.'''
        if not self.cfg:
            return {}
        userPloneGroups = get_plone_groups_for_user()
        reviewProcessInfos = []
        reviewers = self.cfg.reviewersFor()
        for userPloneGroupId in userPloneGroups:
            for reviewer_suffix, review_states in reviewers.items():
                # current user may be able to validate at at least
                # one level of the entire validation process, we take it into account
                if userPloneGroupId.endswith('_%s' % reviewer_suffix):
                    org_uid = userPloneGroupId.split('_')[0]
                    review_states = [
                        '%s%s' % (prefix_review_state, review_state) for review_state in review_states]
                    reviewProcessInfo = [
                        '%s__reviewprocess__%s' % (org_uid, review_state) for review_state in review_states]
                    reviewProcessInfos.extend(reviewProcessInfo)

        if not reviewProcessInfos:
            # in this case, we do not want to display a result
            # we return an unknown review_state
            return _find_nothing_query(self.cfg.getItemTypeName())
        return {'portal_type': {'query': self.cfg.getItemTypeName()},
                'reviewProcessInfo': {'query': reviewProcessInfos}, }


class ItemsToValidateOfMyReviewerGroupsAdapter(BaseItemsToValidateOfMyReviewerGroupsAdapter):

    @property
    @ram.cache(query_user_groups_cachekey)
    def query_itemstovalidateofmyreviewergroups(self):
        return self._query()

    # we may not ram.cache methods in same file with same name...
    query = query_itemstovalidateofmyreviewergroups


class ItemsToCorrectToValidateOfMyReviewerGroupsAdapter(BaseItemsToValidateOfMyReviewerGroupsAdapter):

    @property
    @ram.cache(query_user_groups_cachekey)
    def query_itemstocorrecttovalidateoffmyreviewergroups(self):
        '''Queries all items that current user may correct and in returned_proposed state.'''
        return self._query(prefix_review_state='returned_to_proposing_group_')

    # we may not ram.cache methods in same file with same name...
    query = query_itemstocorrecttovalidateoffmyreviewergroups


class AllItemsToValidateOfMyReviewerGroupsAdapter(BaseItemsToValidateOfMyReviewerGroupsAdapter):

    @property
    @ram.cache(query_user_groups_cachekey)
    def query_allitemstovalidateoffmyreviewergroups(self):
        to_validate_query = self._query()
        to_correct_to_validate_query = self._query(prefix_review_state='returned_to_proposing_group_')
        query = merge_queries([to_validate_query, to_correct_to_validate_query])
        return query

    # we may not ram.cache methods in same file with same name...
    query = query_allitemstovalidateoffmyreviewergroups


class BaseItemsToCorrectAdapter(CompoundCriterionBaseAdapter):

    def _query(self, review_states):
        if not self.cfg:
            return {}
        # for every review_states check what roles are able to edit
        # so we will get groups suffixes linked to these roles and find relevant groups
        wfTool = api.portal.get_tool('portal_workflow')
        itemWF = wfTool.getWorkflowsFor(self.cfg.getItemTypeName())[0]
        reviewProcessInfos = []
        for review_state in review_states:
            if review_state in itemWF.states:
                # roles that may edit
                edit_roles = itemWF.states[review_state].permission_roles[ModifyPortalContent]
                # suffixes information for review_state
                roles_of_suffixes = compute_item_roles_to_assign_to_suffixes(
                    self.cfg, None, review_state)[1]
                # keep suffixes having relevant roles
                suffixes = []
                for suffix, roles in roles_of_suffixes.items():
                    if set(edit_roles).intersection(set(roles)):
                        suffixes.append(suffix)
                # we have suffixes to keep, now find suffixed orgs for current user
                userOrgUids = self.tool.get_orgs_for_user(suffixes=suffixes)
                for userOrgUid in userOrgUids:
                    reviewProcessInfos.append('%s__reviewprocess__%s' % (userOrgUid, review_state))
        if not reviewProcessInfos:
            return _find_nothing_query(self.cfg.getItemTypeName())
        # Create query parameters
        return {'portal_type': {'query': self.cfg.getItemTypeName()},
                'reviewProcessInfo': {'query': reviewProcessInfos}, }


class ItemsToCorrectAdapter(BaseItemsToCorrectAdapter):

    @property
    @ram.cache(query_user_groups_cachekey)
    def query_itemstocorrect(self):
        '''Queries all items that current user may correct.'''
        return self._query(review_states=['returned_to_proposing_group'])

    # we may not ram.cache methods in same file with same name...
    query = query_itemstocorrect


class ItemsToAdviceAdapter(CompoundCriterionBaseAdapter):

    def _query(self, include_hidden_during_redaction=True, only_for_current_user=False):
        '''Queries all items for which the current user must give an advice.'''
        if not self.cfg:
            return {}
        org_uids = self.tool.get_orgs_for_user(suffixes=['advisers'])
        # Consider not_given, asked_again and hidden_during_redaction advices,
        # this search will return 'not delay-aware' and 'delay-aware' advices
        indexAdvisers = [org_uid + '_advice_not_given' for org_uid in org_uids] + \
            ['delay__' + org_uid + '_advice_not_given' for org_uid in org_uids] + \
            [org_uid + '_advice_asked_again' for org_uid in org_uids] + \
            ['delay__' + org_uid + '_advice_asked_again' for org_uid in org_uids]
        if include_hidden_during_redaction:
            indexAdvisers += \
                ['{0}_advice_{1}'.format(org_uid, HIDDEN_DURING_REDACTION_ADVICE_VALUE)
                 for org_uid in org_uids] + \
                ['delay__{0}_advice_{1}'.format(org_uid, HIDDEN_DURING_REDACTION_ADVICE_VALUE)
                 for org_uid in org_uids]
        if only_for_current_user:
            userid = get_current_user_id(self.request)
            tmp_indexAdvisers = list(indexAdvisers)
            indexAdvisers = []
            for v in tmp_indexAdvisers:
                # value that will query element specifically asked to userid
                indexAdvisers.append("{0}__userid__{1}".format(v, userid))
                # query also elements for which advice is asked only to entire
                # advisers groups, so for which no userid was selected
                indexAdvisers.append("{0}__userid__{1}".format(v, ALL_ADVISERS_GROUP_VALUE))
        # Create query parameters
        return {'portal_type': {'query': self.cfg.getItemTypeName()},
                # KeywordIndex 'indexAdvisers' use 'OR' by default
                'indexAdvisers': {'query': indexAdvisers}, }

    @property
    @ram.cache(query_user_groups_cachekey)
    def query_itemstoadvice(self):
        return self._query()

    # we may not ram.cache methods in same file with same name...
    query = query_itemstoadvice


class ItemsToAdviceWithoutHiddenDuringRedactionAdapter(ItemsToAdviceAdapter):

    @property
    @ram.cache(query_user_groups_cachekey)
    def query_itemstoadvicewithouthiddenduringredaction(self):
        return self._query(include_hidden_during_redaction=False)

    # we may not ram.cache methods in same file with same name...
    query = query_itemstoadvicewithouthiddenduringredaction


class MyItemsToAdviceAdapter(ItemsToAdviceAdapter):

    @property
    @ram.cache(query_user_groups_cachekey)
    def query_myitemstoadvice(self):
        return self._query(only_for_current_user=True)

    # we may not ram.cache methods in same file with same name...
    query = query_myitemstoadvice


class MyItemsToAdviceWithoutHiddenDuringRedactionAdapter(ItemsToAdviceAdapter):

    @property
    @ram.cache(query_user_groups_cachekey)
    def query_myitemstoadvicewithouthiddenduringredaction(self):
        return self._query(include_hidden_during_redaction=False, only_for_current_user=True)

    # we may not ram.cache methods in same file with same name...
    query = query_myitemstoadvicewithouthiddenduringredaction


class ItemsToAdviceWithoutDelayAdapter(CompoundCriterionBaseAdapter):

    @property
    @ram.cache(query_user_groups_cachekey)
    def query_itemstoadvicewithoutdelay(self):
        '''Queries all items for which the current user must give an advice without delay.'''
        if not self.cfg:
            return {}
        org_uids = self.tool.get_orgs_for_user(suffixes=['advisers'])
        # Add a '_advice_not_given' at the end of every group id: we want "not given" advices.
        # this search will only return 'not delay-aware' advices
        indexAdvisers = [org_uid + '_advice_not_given' for org_uid in org_uids] + \
            [org_uid + '_advice_asked_again' for org_uid in org_uids] + \
            ['{0}_advice_{1}'.format(org_uid, HIDDEN_DURING_REDACTION_ADVICE_VALUE) for org_uid in org_uids]
        # Create query parameters
        return {'portal_type': {'query': self.cfg.getItemTypeName()},
                # KeywordIndex 'indexAdvisers' use 'OR' by default
                'indexAdvisers': {'query': indexAdvisers}, }

    # we may not ram.cache methods in same file with same name...
    query = query_itemstoadvicewithoutdelay


class ItemsToAdviceWithDelayAdapter(CompoundCriterionBaseAdapter):

    @property
    @ram.cache(query_user_groups_cachekey)
    def query_itemstoadvicewithdelay(self):
        '''Queries all items for which the current user must give an advice with delay.'''
        if not self.cfg:
            return {}
        org_uids = self.tool.get_orgs_for_user(suffixes=['advisers'])
        # Add a '_advice_not_given' at the end of every group id: we want "not given" advices.
        # this search will only return 'delay-aware' advices
        indexAdvisers = ['delay__' + org_uid + '_advice_not_given' for org_uid in org_uids] + \
            ['delay__' + org_uid + '_advice_asked_again' for org_uid in org_uids] + \
            ['delay__{0}_advice_{1}'.format(org_uid, HIDDEN_DURING_REDACTION_ADVICE_VALUE)
             for org_uid in org_uids]
        # Create query parameters
        return {'portal_type': {'query': self.cfg.getItemTypeName()},
                # KeywordIndex 'indexAdvisers' use 'OR' by default
                'indexAdvisers': {'query': indexAdvisers}, }

    # we may not ram.cache methods in same file with same name...
    query = query_itemstoadvicewithdelay


class ItemsToAdviceWithExceededDelayAdapter(CompoundCriterionBaseAdapter):

    @property
    @ram.cache(query_user_groups_cachekey)
    def query_itemstoadvicewithexceededdelay(self):
        '''Queries all items for which the current user must give an advice with exceeded delay.'''
        if not self.cfg:
            return {}
        org_uids = self.tool.get_orgs_for_user(suffixes=['advisers'])
        # Add a '_delay_exceeded' at the end of every group id: we want "not given" advices.
        # this search will only return 'delay-aware' advices for wich delay is exceeded
        indexAdvisers = ['delay__' + org_uid + '_advice_delay_exceeded' for org_uid in org_uids]
        # Create query parameters
        return {'portal_type': {'query': self.cfg.getItemTypeName()},
                # KeywordIndex 'indexAdvisers' use 'OR' by default
                'indexAdvisers': {'query': indexAdvisers}, }

    # we may not ram.cache methods in same file with same name...
    query = query_itemstoadvicewithexceededdelay


class AdvisedItemsAdapter(CompoundCriterionBaseAdapter):

    @property
    @ram.cache(query_user_groups_cachekey)
    def query_adviseditems(self):
        '''Queries items for which an advice has been given.'''
        if not self.cfg:
            return {}
        org_uids = self.tool.get_orgs_for_user(suffixes=['advisers'])
        # advised items are items that has an advice in a particular review_state
        # just append every available meetingadvice state: we want "given" advices.
        # this search will return every advices
        wfTool = api.portal.get_tool('portal_workflow')
        adviceStates = []
        # manage multiple 'meetingadvice' portal_types
        for portal_type_id in getAdvicePortalTypeIds():
            adviceWF = wfTool.getWorkflowsFor(portal_type_id)[0]
            adviceStates += adviceWF.states.keys()
        # remove duplicates
        adviceStates = tuple(set(adviceStates))
        indexAdvisers = []
        for adviceState in adviceStates:
            indexAdvisers += [org_uid + '_%s' % adviceState for org_uid in org_uids]
            indexAdvisers += ['delay__' + org_uid + '_%s' % adviceState for org_uid in org_uids]
        # Create query parameters
        return {'portal_type': {'query': self.cfg.getItemTypeName()},
                # KeywordIndex 'indexAdvisers' use 'OR' by default
                'indexAdvisers': {'query': indexAdvisers}, }

    # we may not ram.cache methods in same file with same name...
    query = query_adviseditems


class AdvisedItemsWithDelayAdapter(CompoundCriterionBaseAdapter):

    @property
    @ram.cache(query_user_groups_cachekey)
    def query_adviseditemswithdelay(self):
        '''Queries items for which an advice has been given with delay.'''
        if not self.cfg:
            return {}
        org_uids = self.tool.get_orgs_for_user(suffixes=['advisers'])
        # advised items are items that has an advice in a particular review_state
        # just append every available meetingadvice state: we want "given" advices.
        # this search will only return 'delay-aware' advices
        wfTool = api.portal.get_tool('portal_workflow')
        adviceStates = []
        # manage multiple 'meetingadvice' portal_types
        for portal_type_id in getAdvicePortalTypeIds():
            adviceWF = wfTool.getWorkflowsFor(portal_type_id)[0]
            adviceStates += adviceWF.states.keys()
        # remove duplicates
        adviceStates = tuple(set(adviceStates))
        indexAdvisers = []
        for adviceState in adviceStates:
            indexAdvisers += ['delay__' + org_uid + '_%s' % adviceState for org_uid in org_uids]
        # Create query parameters
        return {'portal_type': {'query': self.cfg.getItemTypeName()},
                # KeywordIndex 'indexAdvisers' use 'OR' by default
                'indexAdvisers': {'query': indexAdvisers}, }

    # we may not ram.cache methods in same file with same name...
    query = query_adviseditemswithdelay


class DecidedItemsAdapter(CompoundCriterionBaseAdapter):

    @property
    @ram.cache(query_meeting_config_modified_cachekey)
    def query_decideditems(self):
        '''Queries decided items.'''
        if not self.cfg:
            return {}
        return {'portal_type': {'query': self.cfg.getItemTypeName()},
                'review_state': {'query': self.cfg.getItemDecidedStates()}, }

    # we may not ram.cache methods in same file with same name...
    query = query_decideditems


class LivingItemsAdapter(CompoundCriterionBaseAdapter):

    @property
    @ram.cache(query_meeting_config_modified_cachekey)
    def query_livingitems(self):
        '''Queries living items, items not decided yet.'''
        if not self.cfg:
            return {}
        return {'portal_type': {'query': self.cfg.getItemTypeName()},
                'review_state': {'not': self.cfg.getItemDecidedStates()}, }

    # we may not ram.cache methods in same file with same name...
    query = query_livingitems


class PersonalLabelsAdapter(CompoundCriterionBaseAdapter):

    @property
    @ram.cache(query_user_groups_cachekey)
    def query_personal_labels(self):
        '''Special query that will get personal labels defined in DashboardCollection
           query and turn it into ftw.labels compliant personal labels.'''
        if not self.cfg:
            return {}
        # get personal labels to make current user aware and to negativate
        labels = [value for value in self.context.query if value[u'i'] == u'labels']
        if labels:
            member_id = get_current_user_id(self.request)
            # if no selected values, the 'v' key is not there...
            labels = labels[0].get(u'v', [])
            personal_labels = ['{0}:{1}'.format(member_id, label) for label in labels]
        return {'portal_type': {'query': self.cfg.getItemTypeName()},
                'labels': {'query': personal_labels}, }

    # we may not ram.cache methods in same file with same name...
    query = query_personal_labels


class PMNegativePersonalLabelsAdapter(CompoundCriterionBaseAdapter, NegativePersonalLabelsAdapter):

    @property
    @ram.cache(query_user_groups_cachekey)
    def query_negative_personal_labels(self):
        '''Special query that will get personal labels defined in DashboardCollection
           query and turn personal labels to negative personal values.'''
        if not self.cfg:
            return {}
        base_query = super(PMNegativePersonalLabelsAdapter, self).query
        base_query.update({'portal_type': {'query': self.cfg.getItemTypeName()}})
        return base_query

    # we may not ram.cache methods in same file with same name...
    query = query_negative_personal_labels


class PMNegativePreviousIndexValuesAdapter(CompoundCriterionBaseAdapter, NegativePreviousIndexValuesAdapter):

    _adapter_name = u'items-with-negative-previous-index'

    @property
    @ram.cache(query_user_groups_cachekey)
    def query_negative_previous_index_values(self):
        '''Special query that will get previous value in DashboardCollection and
           negativize values.  So for example, if previous index is indexAdvisers,
           values defined in indexAdvisers will be negativized, this will let find
           items that does not have that or that advice.'''
        if not self.cfg:
            return {}

        base_query = super(PMNegativePreviousIndexValuesAdapter, self).query
        base_query.update({'portal_type': {'query': self.cfg.getItemTypeName()}})
        return base_query

    # we may not ram.cache methods in same file with same name...
    query = query_negative_previous_index_values


class SearchItemsOfMyCommitteesAdapter(CompoundCriterionBaseAdapter):

    @property
    @ram.cache(query_user_groups_cachekey)
    def query_itemsofmycommittees(self):
        """Queries all items of committees of the current user."""
        if not self.cfg:
            return {}
        # get every enable committees groups for current MeetingConfig
        # and intersect with user groups
        committee_group_ids = self.cfg.createCommitteeEditorsGroups(
            dry_run_return_group_ids=True)
        user_group_ids = get_plone_groups_for_user()
        committee_user_group_ids = set(committee_group_ids).intersection(
            user_group_ids)
        committee_ids = [committee_user_group_id.split("_", 1)[1]
                         for committee_user_group_id
                         in committee_user_group_ids]
        return {
            "portal_type": {"query": self.cfg.getItemTypeName()},
            "committees_index": {"query": sorted(committee_ids)},
        }

    # we may not ram.cache methods in same file with same name...
    query = query_itemsofmycommittees


class SearchItemsOfMyCommitteesEditableAdapter(SearchItemsOfMyCommitteesAdapter):

    @property
    @ram.cache(query_user_groups_cachekey)
    def query_itemsofmycommitteeseditable(self):
        """Queries all items of committees of the current user."""
        if not self.cfg:
            return {}
        # this manage "portal_type" and "committees_index"
        base_query = self.query_itemsofmycommittees
        # complete with review_states in which items committees fields are editable
        base_query.update({"review_state": {"query": self.cfg.getItemCommitteesStates()}})
        return base_query

    # we may not ram.cache methods in same file with same name...
    query = query_itemsofmycommitteeseditable


class PMCategorizedObjectInfoAdapter(CategorizedObjectInfoAdapter):
    """ """

    def __init__(self, context):
        super(PMCategorizedObjectInfoAdapter, self).__init__(context)
        self.tool = api.portal.get_tool('portal_plonemeeting')
        self.cfg = self.tool.getMeetingConfig(context)
        self.parent = self.context.aq_inner.aq_parent

    def get_infos(self, category, limited=False):
        """A the 'visible_for_groups' info."""
        if not limited:
            visible_for_groups = self._visible_for_groups()
        infos = super(PMCategorizedObjectInfoAdapter, self).get_infos(
            category, limited=limited)
        if not limited:
            infos['visible_for_groups'] = visible_for_groups
        return infos

    def _apply_visible_groups_security(self, group_ids):
        """Compute 'View' permission if annex is confidential,
           apply local_roles and give 'View' to 'AnnexReader' either,
           remove every local_roles and acquire 'View'.
           WE DO NOT apply this on Meeting because of the possibility to select
           "every _creators groups" and it is not possible to give the
           'AnnexReader' role to all these _creators groups."""
        if self.parent.getTagName() == 'MeetingItem' or \
           self.parent.portal_type in getAdvicePortalTypeIds():
            # reinitialize permissions in case no more confidential
            # or confidentiality configuration changed
            self.context.__ac_local_roles_block__ = False
            self.context.manage_permission(View, (), acquire=True)
            self.context.manage_permission(AccessContentsInformation, (), acquire=True)
            grp_reader_localroles = [
                grp_id for grp_id in self.context.__ac_local_roles__
                if READER_USECASES['confidentialannex'] in self.context.__ac_local_roles__[grp_id]]
            self.context.manage_delLocalRoles(grp_reader_localroles)
            if self.context.confidential:
                self.context.manage_permission(
                    View,
                    (READER_USECASES['confidentialannex'], 'Manager', 'MeetingManager'),
                    acquire=False)
                self.context.manage_permission(
                    AccessContentsInformation,
                    (READER_USECASES['confidentialannex'], 'Manager', 'MeetingManager'),
                    acquire=False)
                for grp_id in group_ids:
                    self.context.manage_addLocalRoles(
                        grp_id, (READER_USECASES['confidentialannex'], ))
            # self.context.reindexObjectSecurity()

    def _visible_for_groups(self):
        """ """
        groups = []
        confidential = getattr(self.context, 'confidential', False)
        if confidential:
            groups = self._compute_visible_for_groups()

        # apply security if confidential or going from confidential to not confidential
        # in this case, the 'View' permission was not acquired
        if confidential or not self.context.acquiredRolesAreUsedBy('View'):
            self._apply_visible_groups_security(groups)
        return groups

    @memoize
    def _compute_visible_for_groups(self):
        """ """
        groups = []
        parent_classname = self.parent.getTagName()
        if parent_classname == 'MeetingItem':
            visible_fors = self.cfg.getItemAnnexConfidentialVisibleFor()
            groups = self._item_visible_for_groups(visible_fors, item=self.parent)
        elif parent_classname == 'Meeting':
            visible_fors = self.cfg.getMeetingAnnexConfidentialVisibleFor()
            groups = self._meeting_visible_for_groups(visible_fors)
        else:
            # advice
            visible_fors = self.cfg.getAdviceAnnexConfidentialVisibleFor()
            groups = self._advice_visible_for_groups(
                visible_fors, item=self.parent.aq_parent)
        return groups

    def _item_visible_for_groups(self, visible_fors, item):
        """ """
        res = []
        res += self._configgroup_groups(visible_fors)
        res += self._reader_groups(visible_fors, item)
        res += self._suffix_proposinggroup(visible_fors, item)
        return res

    def _meeting_visible_for_groups(self, visible_fors):
        """ """
        res = []
        res += self._configgroup_groups(visible_fors)
        res += self._suffix_profile_proposinggroup(visible_fors)
        return res

    def _advice_visible_for_groups(self, visible_fors, item):
        """ """
        res = []
        res += self._configgroup_groups(visible_fors)
        res += self._reader_groups(visible_fors, item)
        res += self._suffix_proposinggroup(visible_fors, item)
        if 'adviser_group' in visible_fors:
            plone_group_id = get_plone_group_id(self.parent.advice_group, 'advisers')
            res.append(plone_group_id)
        return res

    def _configgroup_groups(self, visible_fors):
        """ """
        res = []
        for visible_for in visible_fors:
            if visible_for.startswith(CONFIGGROUPPREFIX):
                suffix = visible_for.replace(CONFIGGROUPPREFIX, '')
                res.append('{0}_{1}'.format(self.cfg.getId(), suffix))
        return res

    def _suffix_proposinggroup(self, visible_fors, item=None):
        """Behavior of this method change when receiving an item or not:
           - when p_item is not None, we will compute Plone groups of p_visible_fors
           suffixes of groups managing the item;
           - when p_item is None, we will consider every groups using the suffixes."""
        res = []
        # item, we take managing groups
        if item:
            org_uids = item.adapted()._getAllGroupsManagingItem(
                item.query_state())
        else:
            # every enabled groups
            org_uids = get_organizations(the_objects=False)
        for visible_for in visible_fors:
            if visible_for.startswith(PROPOSINGGROUPPREFIX):
                suffix = visible_for.replace(PROPOSINGGROUPPREFIX, '')
                for org_uid in org_uids:
                    plone_group_id = get_plone_group_id(org_uid, suffix)
                    res.append(plone_group_id)
        return res

    def _suffix_profile_proposinggroup(self, visible_fors):
        """ """
        res = []
        for visible_for in visible_fors:
            if visible_for.startswith(SUFFIXPROFILEPREFIX):
                res.append(visible_for)
        return res

    def _reader_groups(self, visible_fors, item=None):
        """ """
        res = []
        for visible_for in visible_fors:
            if visible_for == '{0}advices'.format(READERPREFIX):
                # item advisers if item or every possible advisers
                if item:
                    org_uids = item.adviceIndex.keys()
                else:
                    # every possible advisers, so configured custom advisers and selectable advisers
                    custom_advisers_org_uids = [row['org_uid'] for row in self.cfg.getCustomAdvisers()]
                    selectable_advisers = self.cfg.getSelectableAdvisers()
                    org_uids = set(custom_advisers_org_uids).union(selectable_advisers)
                for org_uid in org_uids:
                    plone_group_id = get_plone_group_id(org_uid, 'advisers')
                    res.append(plone_group_id)
            elif visible_for == '{0}copy_groups'.format(READERPREFIX):
                # item copyGroups if item or every possible copy groups
                if item:
                    res = res + list(item.getAllCopyGroups(auto_real_plone_group_ids=True))
                else:
                    res += list(self.cfg.getSelectableCopyGroups())
            elif visible_for == '{0}groupsincharge'.format(READERPREFIX):
                # item groupsInCharges if item or every possible groups in charge
                if item:
                    org_uids = item.getGroupsInCharge(theObjects=False, includeAuto=True)
                else:
                    org_uids = get_vocab_values(
                        self.cfg,
                        "Products.PloneMeeting.vocabularies.groupsinchargevocabulary")
                for org_uid in org_uids:
                    plone_group_id = get_plone_group_id(org_uid, 'observers')
                    res.append(plone_group_id)
        return res


class PMCategorizedObjectAdapter(CategorizedObjectAdapter):
    """ """

    def __init__(self, context, request, categorized_obj):
        super(PMCategorizedObjectAdapter, self).__init__(context, request, categorized_obj)
        self.tool = api.portal.get_tool('portal_plonemeeting')
        self.cfg = self.tool.getMeetingConfig(self.context)

    def can_view(self):
        # as we let user access a not viewable annex, check that it is not an anonymous!
        if api.user.is_anonymous():
            return
        # base adapter check View permission of categorized object
        can_view = super(PMCategorizedObjectAdapter, self).can_view()
        infos = self.context.categorized_elements[self.categorized_obj.UID()]
        class_name = self.context.__class__.__name__
        # special management for not confidential annexes displayed on not viewable context
        if not can_view and not infos['confidential']:
            # check if displaying annexes on a not viewable item using the linked items on an item
            # if not viewable annexes are actually displayed,
            # check that current user has access to a viewable linked item
            if (class_name == 'MeetingItem' and
                'MeetingItem.annexes' in self.cfg.getItemsNotViewableVisibleFields()) and \
               (self.context.adapted().get_predecessors(only_viewable=True) or
                    self.context.getManuallyLinkedItems(only_viewable=True)):
                return True
            # displaying annexes on an inheritated advice when
            # original advice holder is not viewable
            if class_name == 'MeetingAdvice':
                # depending on called view, context will change :
                # @@categorized-childs context is annex of original advice,
                # we can get PUBLISHED object
                # @@categorized-childs-infos or @@download context is an annex,
                # we get current context using the referer
                item = get_context_with_request(None)
                if item.__class__.__name__ != 'MeetingItem':
                    item = get_referer_obj(self.request)
                if item.adviceIndex[self.context.advice_group]['inherited'] is True and \
                   _checkPermission(View, item):
                    return True
        else:
            # is the context a MeetingItem and privacy viewable?
            if class_name == 'MeetingItem' and \
               self.cfg.getRestrictAccessToSecretItems() and \
               not self.context.adapted().isPrivacyViewable():
                return False

            # bypass if not confidential
            if not infos['confidential']:
                return True

            # bypass for MeetingManagers
            if self.tool.isManager(self.cfg):
                return True

            # Meeting
            if class_name == 'Meeting':
                # if we have a SUFFIXPROFILEPREFIX prefixed group,
                # check using "userIsAmong", this is only done for Meetings
                if set(get_plone_groups_for_user()).intersection(
                        infos['visible_for_groups']):
                    return True
                # build suffixes to pass to tool.userIsAmong
                suffixes = []
                for group in infos['visible_for_groups']:
                    if group.startswith(SUFFIXPROFILEPREFIX):
                        suffixes.append(group.replace(SUFFIXPROFILEPREFIX, ''))
                if suffixes and self.tool.userIsAmong(suffixes, cfg=self.cfg):
                    return True
                return False

            return can_view


class IconifiedCategoryConfigAdapter(object):
    """ """
    def __init__(self, context):
        """ """
        self.context = context

    @memoize
    def get_config(self):
        """ """
        tool = api.portal.get_tool('portal_plonemeeting')
        # manage the css.py file generation necessary CSS, in this case, context is the portal
        # we return portal as the config root so css file is generated with every existing categories
        # found in every MeetingConfigs
        if self.context.portal_type == 'Plone Site':
            return self.context
        try:
            cfg = tool.getMeetingConfig(self.context)
        except AttributeError:
            cfg = None
        return cfg and cfg.annexes_types or cfg


class IconifiedCategoryGroupAdapter(object):
    """ """
    def __init__(self, config, context):
        """ """
        self.config = config
        self.context = context
        self.request = getattr(self.context, 'REQUEST', {})

    @memoize
    def get_group(self):
        """Return right group, depends on :
           - while adding in an item, annex or decisionAnnex;
           - while adding in a meeting or an advice."""
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(self.context)
        parent = self.context.getParentNode()
        # adding annex to an item
        if self.context.getTagName() == 'MeetingItem' or \
           (self.context.portal_type in ('annex', 'annexDecision') and
                parent.getTagName() == 'MeetingItem'):
            isItemDecisionAnnex = False
            if self.context.getTagName() == 'MeetingItem':
                # it is possible to force to use the item_decision_annexes group
                # or when using quickupload, the typeupload contains the type of element to add
                if self.request.get('force_use_item_decision_annexes_group', False) or \
                   self.request.get('typeupload', None) == 'annexDecision':
                    return cfg.annexes_types.item_decision_annexes

                # we are adding a new annex, get annex portal_type from form_instance
                # manage also the InlineValidation view
                if hasattr(self.request.get('PUBLISHED'), 'form_instance'):
                    form_instance = self.request.get('PUBLISHED').form_instance
                elif (hasattr(self.request.get('PUBLISHED'), 'context',) and
                      hasattr(self.request.get('PUBLISHED').context, 'form_instance')):
                    form_instance = self.request.get('PUBLISHED').context.form_instance
                else:
                    # calling with MeetingItem as context, this is the case when checking
                    # if categories exist and if annexes tab should be displayed
                    return cfg.annexes_types.item_annexes

                if getattr(form_instance, 'portal_type', '') == 'annexDecision':
                    isItemDecisionAnnex = True
            else:
                if self.context.portal_type == 'annexDecision':
                    isItemDecisionAnnex = True

            if not isItemDecisionAnnex:
                return cfg.annexes_types.item_annexes
            else:
                return cfg.annexes_types.item_decision_annexes

        # adding annex to an advice
        if self.context.getTagName() == "MeetingAdvice" or \
           parent.getTagName() == "MeetingAdvice":
            return cfg.annexes_types.advice_annexes

        # adding annex to a meeting
        if self.context.getTagName() == 'Meeting' or \
           parent.getTagName() == 'Meeting':
            return cfg.annexes_types.meeting_annexes

    def get_every_categories(self, only_enabled=True):
        categories = get_categories(
            self.context, the_objects=True, only_enabled=only_enabled)
        self.request['force_use_item_decision_annexes_group'] = True
        categories = categories + get_categories(
            self.context, the_objects=True, only_enabled=only_enabled)
        self.request['force_use_item_decision_annexes_group'] = False
        return categories


class PMGenerablePODTemplatesAdapter(GenerablePODTemplatesAdapter):
    """ """

    def get_all_pod_templates(self):
        """Query by MeetingConfig."""
        try:
            tool = api.portal.get_tool('portal_plonemeeting')
        except InvalidParameterError:
            return []

        cfg = tool.getMeetingConfig(self.context)
        if not cfg:
            return []
        # do not render on other contents than IMeetingContent
        # or it is also computed on Folder (dashboards) and other Plone pages
        if not IMeetingContent.providedBy(self.context):
            return []

        # OK, we are on a IMeetingContent, compute the pod templates
        catalog = api.portal.get_tool('portal_catalog')
        brains = catalog.unrestrictedSearchResults(
            portal_type='ConfigurablePODTemplate',
            getConfigId=cfg.getId(),
            enabled=True,
            sort_on='getObjPositionInParent'
        )
        pod_templates = [self.context.unrestrictedTraverse(brain.getPath()) for brain in brains]

        return pod_templates


class PMDashboardGenerablePODTemplatesAdapter(DashboardGenerablePODTemplatesAdapter):
    """ """

    def get_all_pod_templates(self):
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(self.context)
        query = {'object_provides': {'query': IDashboardPODTemplate.__identifier__},
                 'enabled': True,
                 'sort_on': 'getObjPositionInParent'}
        # filter on MeetingConfig if we are in it
        if cfg:
            query['getConfigId'] = cfg.getId()
        else:
            # out of a MeetingConfig
            query['getConfigId'] = EMPTY_STRING

        catalog = api.portal.get_tool('portal_catalog')
        brains = catalog.unrestrictedSearchResults(**query)
        pod_templates = [self.context.unrestrictedTraverse(brain.getPath()) for brain in brains]

        return pod_templates


#########################
# vocabularies adapters #
#########################


class AnnexMissingTermsVocabulary(MissingChoiceTermsVocabulary, MissingTerms):
    """ Managing missing terms for IAnnex. """

    def complete_voc(self):
        if self.field.getName() == 'content_category':
            return get_vocab(self.context,
                             'collective.iconifiedcategory.every_categories')
        else:
            return SimpleVocabulary([])



# -*- coding: utf-8 -*-
#
# File: testVotes.py
#
# GNU General Public License (GPL)
#

from AccessControl import Unauthorized
from imio.helpers.cache import cleanRamCache
from plone import api
from Products.PloneMeeting.browser.itemvotes import _should_disable_apply_until_item_number
from Products.PloneMeeting.browser.itemvotes import IEncodeSecretVotes
from Products.PloneMeeting.browser.itemvotes import secret_votes_default
from Products.PloneMeeting.browser.itemvotes import votes_default
from Products.PloneMeeting.config import ALL_VOTE_VALUES
from Products.PloneMeeting.config import NOT_ENCODED_VOTE_VALUE
from Products.PloneMeeting.config import NOT_VOTABLE_LINKED_TO_VALUE
from Products.PloneMeeting.content.meeting import IMeeting
from Products.PloneMeeting.tests.PloneMeetingTestCase import PloneMeetingTestCase
from Products.PloneMeeting.utils import may_view_field
from Products.statusmessages.interfaces import IStatusMessage
from z3c.form import validator
from zope.i18n import translate
from zope.interface import Invalid

import transaction


class testVotes(PloneMeetingTestCase):
    '''Tests various aspects of votes management.'''

    def setUp(self):
        # call parent setUp
        super(testVotes, self).setUp()
        self._setUpOrderedContacts()
        self._removeConfigObjectsFor(self.meetingConfig)

    def _createMeetingWithVotes(self, include_yes=True, include_no_vote=False):
        """ """
        self.changeUser('pmManager')
        meeting = self.create('Meeting')
        public_item = self.create('MeetingItem',
                                  decision=self.decisionText)
        secret_item = self.create('MeetingItem',
                                  decision=self.decisionText,
                                  pollType='secret')
        self.presentItem(public_item)
        if include_yes:
            yes_public_item = self.create('MeetingItem',
                                          decision=self.decisionText)
            self.presentItem(yes_public_item)
        self.presentItem(secret_item)
        if include_yes:
            yes_secret_item = self.create('MeetingItem',
                                          decision=self.decisionText,
                                          pollType='secret')
            self.presentItem(yes_secret_item)
        if include_no_vote:
            no_vote_item = self.create('MeetingItem',
                                       decision=self.decisionText,
                                       pollType='no_vote')
            self.presentItem(no_vote_item)
        voters = meeting.get_voters()
        # public votes
        public_votes = public_item.get_item_votes()[0]
        public_votes['voters'][voters[0]] = "yes"
        public_votes['voters'][voters[1]] = "yes"
        public_votes['voters'][voters[2]] = "no"
        public_votes['voters'][voters[3]] = "abstain"
        meeting.set_item_public_vote(public_item, public_votes, 0)
        transaction.commit()
        # encode secret votes
        secret_votes = secret_item.get_item_votes()[0]
        secret_votes['votes']['abstain'] = 2
        secret_votes['votes']['no'] = 1
        secret_votes['votes']['yes'] = 1
        meeting.set_item_secret_vote(secret_item, secret_votes, 0)
        transaction.commit()
        # all yes public votes
        if include_yes:
            public_votes['voters'][voters[2]] = "yes"
            public_votes['voters'][voters[3]] = "yes"
            meeting.set_item_public_vote(yes_public_item, public_votes, 0)
            transaction.commit()
            # all yes secret votes
            secret_votes['votes']['abstain'] = 0
            secret_votes['votes']['no'] = 0
            secret_votes['votes']['yes'] = 4
            meeting.set_item_secret_vote(yes_secret_item, secret_votes, 0)
            transaction.commit()
        res = meeting, public_item, secret_item
        if include_yes:
            res = meeting, public_item, yes_public_item, secret_item, yes_secret_item
        if include_no_vote:
            res += (no_vote_item, )
        return res

    def test_pm_Show_votes(self):
        """Votes are only shown on an item presented to a meeting,
           unless pollType is "no_vote"."""
        cfg = self.meetingConfig
        self.changeUser('pmManager')
        self.assertTrue(cfg.getUseVotes())
        meeting = self.create('Meeting')
        self.assertTrue(meeting.get_voters())
        item = self.create('MeetingItem')
        self.assertFalse(item.show_votes())
        self.presentItem(item)
        self.assertEqual(item.getPollType(), 'freehand')
        self.assertTrue(item.show_votes())
        item.setPollType('secret')
        self.assertTrue(item.show_votes())
        item.setPollType('no_vote')
        self.assertFalse(item.show_votes())
        item.setPollType('secret')
        self.assertTrue(item.show_votes())
        # disable votes
        cfg.setUseVotes(False)
        # still shown because voters
        self.assertTrue(item.show_votes())

        # do not show_votes if not voter
        # this will avoid showing votes on older
        # meeting where votes were not enabled
        no_vote_meeting = self.create('Meeting')
        no_vote_item = self.create('MeetingItem')
        self.presentItem(no_vote_item)
        self.assertFalse(no_vote_meeting.get_voters())
        self.assertFalse(no_vote_item.get_item_voters())
        self.assertFalse(no_vote_item.show_votes())
        # even if enabled, if nothing to show, nothing shown
        cfg.setUseVotes(True)
        self.assertFalse(no_vote_item.show_votes())

    def test_pm_GetItemVotes(self):
        """Returns votes on an item."""
        self.changeUser('pmManager')
        public_item = self.create('MeetingItem')
        secret_item = self.create('MeetingItem', pollType='secret')
        meeting = self.create('Meeting')
        # return an empty list of not linked to a meeting
        self.assertEqual(public_item.get_item_votes(), [])
        self.assertEqual(secret_item.get_item_votes(), [])
        self.presentItem(public_item)
        self.presentItem(secret_item)
        # return an empty vote when nothing encoded
        # when include_unexisting=True (default)
        public_vote = public_item.get_item_votes()[0]
        secret_vote = secret_item.get_item_votes()[0]
        self.assertEqual(public_vote['vote_number'], 0)
        self.assertEqual(secret_vote['vote_number'], 0)
        # voters are on public vote
        voters = meeting.get_voters()
        self.assertEqual(sorted(public_vote['voters'].keys()), sorted(voters))
        # not on secret
        self.assertFalse('voters' in secret_vote)
        self.assertTrue('votes' in secret_vote)
        self.assertTrue('yes' in secret_vote['votes'])
        self.assertTrue('no' in secret_vote['votes'])
        self.assertTrue('abstain' in secret_vote['votes'])

    def test_pm_PrintVotes(self):
        """Test the print_votes helper."""
        meeting, public_item, yes_public_item, secret_item, yes_secret_item = \
            self._createMeetingWithVotes()

        # print_votes
        view = public_item.restrictedTraverse('document-generation')
        helper_public = view.get_generation_context_helper()
        view = yes_public_item.restrictedTraverse('document-generation')
        helper_yes_public = view.get_generation_context_helper()
        view = secret_item.restrictedTraverse('document-generation')
        helper_secret = view.get_generation_context_helper()
        view = yes_secret_item.restrictedTraverse('document-generation')
        helper_yes_secret = view.get_generation_context_helper()
        # test also total_voters using include_total_voters=True
        # public vote
        self.assertEqual(
            helper_public.print_votes(include_total_voters=True),
            u'<p>Il y a 4 votants.</p><p>Par 2 voix pour, une voix contre et une abstention,</p>')
        self.assertEqual(
            helper_public.print_votes(single_vote_value=u"1", no_votes_marker="<!>"),
            u'<p>Par 2 voix pour, 1 voix contre et 1 abstention,</p>')
        # is_all_count is also available on the helper view
        self.assertFalse(helper_public.is_all_count())
        self.assertTrue(helper_yes_public.is_all_count())
        # public vote all yes
        self.assertEqual(
            helper_yes_public.print_votes(include_total_voters=True),
            u"<p>Il y a 4 votants.</p><p>\xc0 l'unanimit\xe9,</p>")
        # secret vote
        self.assertEqual(helper_secret.print_votes(include_total_voters=True),
                         u'<p>Il y a 4 votants.</p><p>Au scrutin secret,</p>'
                         u'<p>Par une voix pour, une voix contre et 2 abstentions,</p>')
        # secret vote all yes and secret_intro
        self.assertEqual(helper_yes_secret.print_votes(secret_intro=u"<p>À bulletin secret,</p>"),
                         u"<p>\xc0 bulletin secret,</p>"
                         u"<p>\xc0 l'unanimit\xe9,</p>")

        # include_voters=True
        # public vote
        self.assertEqual(
            helper_public.print_votes(include_voters=True),
            u'<p>Par 2 voix pour<p>Monsieur Person1FirstName Person1LastName, '
            u'Assembly member 1, Monsieur Person2FirstName Person2LastName, '
            u'Assembly member 2</p>, une voix contre<p>Madame Person3FirstName '
            u'Person3LastName, Assembly member 3</p> et une abstention<p>Madame '
            u'Person4FirstName Person4LastName, Assembly member 4 &amp; 5</p>,</p>')
        # include_voters=('no', 'absention')
        self.assertEqual(
            helper_public.print_votes(include_voters=('no', 'abstain')),
            u'<p>Par 2 voix pour, une voix contre'
            u'<p>Madame Person3FirstName Person3LastName, Assembly member 3</p> '
            u'et une abstention<p>Madame Person4FirstName Person4LastName, '
            u'Assembly member 4 &amp; 5</p>,</p>')
        # include_voters_percent_treshold=40
        # will display voters for vote values where there are less than 40% of the total voters
        self.assertEqual(
            helper_public.print_votes(include_voters=True, include_voters_percent_treshold=40),
            u'<p>Par 2 voix pour, une voix contre'
            u'<p>Madame Person3FirstName Person3LastName, Assembly member 3</p> '
            u'et une abstention<p>Madame Person4FirstName Person4LastName, '
            u'Assembly member 4 &amp; 5</p>,</p>')
        # public vote all yes
        self.assertEqual(
            helper_yes_public.print_votes(include_voters=True),
            u"<p>\xc0 l'unanimit\xe9,</p>")
        self.assertEqual(
            helper_yes_public.print_votes(include_voters=True, all_yes_render=None),
            u'<p>Par 4 voix pour<p>'
            u'Monsieur Person1FirstName Person1LastName, Assembly member 1, '
            u'Monsieur Person2FirstName Person2LastName, Assembly member 2, '
            u'Madame Person3FirstName Person3LastName, Assembly member 3, '
            u'Madame Person4FirstName Person4LastName, Assembly member 4 &amp; 5</p>,</p>')
        # change an assembly member order, it is taken into account
        change_view = yes_public_item.restrictedTraverse('@@item-change-attendee-order')
        change_view(attendee_uid=public_item.get_all_attendees()[0], position=3)
        self.assertEqual(
            helper_yes_public.print_votes(include_voters=True, all_yes_render=None),
            u'<p>Par 4 voix pour'
            u'<p>Monsieur Person2FirstName Person2LastName, Assembly member 2, '
            u'Madame Person3FirstName Person3LastName, Assembly member 3, '
            u'Monsieur Person1FirstName Person1LastName, Assembly member 1, '
            u'Madame Person4FirstName Person4LastName, Assembly member 4 &amp; 5</p>,</p>')

        # other possible vote values
        cfg = self.meetingConfig
        cfg.setUsedVoteValues(ALL_VOTE_VALUES)

        voters = meeting.get_voters()
        # public votes
        public_votes = public_item.get_item_votes()[0]
        public_votes['voters'][voters[0]] = "blank"
        public_votes['voters'][voters[1]] = "invalid"
        public_votes['voters'][voters[2]] = "does_not_vote"
        public_votes['voters'][voters[3]] = "not_found"
        meeting.set_item_public_vote(public_item, public_votes, 0)
        transaction.commit()
        # encode secret votes
        secret_votes = secret_item.get_item_votes()[0]
        secret_votes['votes']['blank'] = 1
        secret_votes['votes']['invalid'] = 1
        secret_votes['votes']['does_not_vote'] = 1
        secret_votes['votes']['not_found'] = 1
        meeting.set_item_secret_vote(secret_item, secret_votes, 0)
        transaction.commit()
        self.assertEqual(
            helper_public.print_votes(),
            u'<p>Par un bulletin "ne vote pas", un bulletin non trouv\xe9 dans l\'urne, '
            u'un bulletin invalide et un vote blanc,</p>')
        # if single_vote_value is uncomplete dict, fallback to "1"
        self.assertEqual(
            helper_public.print_votes(single_vote_value={'yes': 'one'}),
            u'<p>Par 1 bulletin "ne vote pas", 1 bulletin non trouv\xe9 dans l\'urne, '
            u'1 bulletin invalide et 1 vote blanc,</p>')
        self.assertEqual(
            helper_public.print_votes(used_patterns="counts", single_vote_value="1"),
            u'<p>Par <p><strong>Ne vote pas: 1</strong></p>, '
            u'<p><strong>Bulletin non trouv\xe9: 1</strong></p>, '
            u'<p><strong>Bulletin invalide: 1</strong></p> et '
            u'<p><strong>Vote blanc: 1</strong></p>,</p>')
        self.assertEqual(
            helper_public.print_votes(used_patterns="counts_persons", single_vote_value="1"),
            u"<p>Par <p><strong>N'a pas vot\xe9: 1</strong></p>, "
            u"<p><strong>Bulletin non trouv\xe9: 1</strong></p>, "
            u"<p><strong>Bulletin invalide: 1</strong></p> et "
            u"<p><strong>A vot\xe9 blanc: 1</strong></p>,</p>")

        # vote_label_pattern
        public_votes['label'] = "My label|My second label|My third label"
        meeting.set_item_public_vote(public_item, public_votes, 0)
        transaction.commit()
        # just render label
        self.assertEqual(
            helper_public.print_votes(
                main_pattern="",
                used_vote_values=[''],
                vote_label_pattern=u"<p>{0}</p>"),
            u'<p>My label</p>')
        self.assertEqual(
            helper_public.print_votes(
                main_pattern="",
                used_vote_values=[''],
                vote_label_pattern=u"<p>{0}: {1}</p>"),
            u'<p>My label: My second label</p>')
        self.assertEqual(
            helper_public.print_votes(
                main_pattern="",
                used_vote_values=[''],
                vote_label_pattern=u"<p>{0}: {2}</p>"),
            u'<p>My label: My third label</p>')

        # no votes
        meeting.item_votes[public_item.UID()] = []
        self.assertEqual(helper_public.print_votes(no_votes_marker="-"), "-")
        self.assertEqual(helper_public.print_votes(no_votes_marker="There is no votes"),
                         "There is no votes")
        self.assertEqual(helper_public.print_votes(no_votes_marker="Aucun vote encodé."),
                         "Aucun vote encodé.")
        self.assertEqual(helper_public.print_votes(no_votes_marker=""), "")

    def test_pm_ItemDeleteVoteView(self):
        """This view will remove a vote, only doable by MeetingManagers."""
        self.changeUser('pmManager')
        meeting, public_item, yes_public_item, secret_item, yes_secret_item = \
            self._createMeetingWithVotes()

        # public_item
        self.changeUser('pmCreator1')
        delete_view = public_item.restrictedTraverse('@@item_delete_vote')
        self.assertRaises(Unauthorized, delete_view, 0, redirect=False)
        self.changeUser('pmManager')
        self.assertTrue(delete_view.context.get_item_votes(include_unexisting=False))
        delete_view(0, redirect=False)
        self.assertFalse(delete_view.context.get_item_votes(include_unexisting=False))

        # secret_item
        self.changeUser('pmCreator1')
        delete_view = secret_item.restrictedTraverse('@@item_delete_vote')
        self.assertRaises(Unauthorized, delete_view, 0, redirect=False)
        self.changeUser('pmManager')
        self.assertTrue(delete_view.context.get_item_votes(include_unexisting=False))
        delete_view(0, redirect=False)
        self.assertFalse(delete_view.context.get_item_votes(include_unexisting=False))

    def test_pm_ItemDeleteVoteViewCanNotDeleteFirstLinkedVote(self):
        """When votes are linked, the first linked vote may not be deleted."""
        self.changeUser('pmManager')
        meeting, public_item, yes_public_item, secret_item, yes_secret_item = \
            self._createMeetingWithVotes()

        # set linked votes
        self.changeUser('pmManager')
        public_votes = public_item.get_item_votes()[0]
        meeting.set_item_public_vote(public_item, public_votes, 0)
        transaction.commit()
        public_votes['linked_to_previous'] = True
        meeting.set_item_public_vote(public_item, public_votes, 1)
        transaction.commit()
        self.assertEqual(len(public_item.get_item_votes(include_unexisting=False)), 2)
        # vote 0 is not deletable
        self.assertFalse(public_item._voteIsDeletable(meeting, 0))
        self.assertTrue(public_item._voteIsDeletable(meeting, 1))
        delete_view = public_item.restrictedTraverse('@@item_delete_vote')
        self.assertRaises(AssertionError, delete_view, object_uid=0)
        # delete vote 1, then vote 0 is deletable
        delete_view(object_uid=1)
        self.assertTrue(public_item._voteIsDeletable(meeting, 0))
        delete_view(object_uid=0)
        self.assertFalse(public_item.get_item_votes(include_unexisting=False))

    def test_pm_ItemDeleteVoteViewDeleteSeveralNotLinkedVotes(self):
        """When votes are not linked, any may be deleted."""
        self.changeUser('pmManager')
        meeting, public_item, yes_public_item, secret_item, yes_secret_item = \
            self._createMeetingWithVotes()

        # add several votes
        self.changeUser('pmManager')
        public_votes = public_item.get_item_votes()[0]
        meeting.set_item_public_vote(public_item, public_votes, 0)
        transaction.commit()
        meeting.set_item_public_vote(public_item, public_votes, 1)
        transaction.commit()
        meeting.set_item_public_vote(public_item, public_votes, 2)
        transaction.commit()
        self.assertEqual(len(public_item.get_item_votes(include_unexisting=False)), 3)
        self.assertTrue(public_item._voteIsDeletable(meeting, 0))
        self.assertTrue(public_item._voteIsDeletable(meeting, 1))
        self.assertTrue(public_item._voteIsDeletable(meeting, 2))
        delete_view = public_item.restrictedTraverse('@@item_delete_vote')
        delete_view(object_uid=1)
        # vote 2 is not vote 1
        delete_view(object_uid=1)
        delete_view(object_uid=0)

    def test_pm_CanNotUnselectVoterOnMeetingIfUsedOnItem(self):
        """This will not be possible to unselect a voter on a meeting
           if it voted on an item :
           - either public vote;
           - or secret vote (number of voters).
        """
        self.changeUser('pmManager')
        meeting, public_item, yes_public_item, secret_item, yes_secret_item = \
            self._createMeetingWithVotes()

        attendee_uids = meeting.get_attendees()
        # now while validating meeting_attendees, None may be unselected
        meeting_attendees = ['muser_{0}_attendee'.format(attendee_uid)
                             for attendee_uid in attendee_uids]
        meeting_voters = ['muser_{0}'.format(attendee_uid)
                          for attendee_uid in attendee_uids]

        # now test with meeting_attendees
        self.request.form['meeting_attendees'] = meeting_attendees
        self.request.form['meeting_voters'] = meeting_voters
        self.assertEqual(meeting.validate(self.request), {})

        # unselecting one would break validation
        # public
        invariants = validator.InvariantsValidator(None, None, None, IMeeting, None)
        self.request.set('validate_dates_done', True)
        voter0 = meeting_voters.pop(0)
        self.request.form['meeting_voters'] = meeting_voters
        public_error_msg = translate(
            u'can_not_remove_public_voter_voted_on_items',
            domain='PloneMeeting',
            mapping={
                'attendee_title':
                    u'Monsieur Person1FirstName Person1LastName, '
                    u'Assembly member 1 (Mon organisation)'},
            context=self.request)
        data = {}
        edit_form = meeting.restrictedTraverse('@@edit')
        edit_form.update()
        self.request['PUBLISHED'] = edit_form
        errors = invariants.validate(data)
        self.request.set('validate_attendees_done', False)
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].message, public_error_msg)
        meeting_voters.insert(0, voter0)
        self.assertEqual(invariants.validate(data), ())
        self.request.set('validate_attendees_done', False)

        # secret
        voter0 = meeting_voters.pop(0)
        # remove public votes
        delete_view = public_item.restrictedTraverse('@@item_delete_vote')
        delete_view(0, redirect=False)
        self.assertEqual(public_item.get_item_votes(include_unexisting=False), [])
        delete_view = yes_public_item.restrictedTraverse('@@item_delete_vote')
        delete_view(0, redirect=False)
        self.assertEqual(yes_public_item.get_item_votes(include_unexisting=False), [])
        self.request.form['meeting_voters'] = meeting_voters
        secret_error_msg = translate(
            u'can_not_remove_secret_voter_voted_on_items',
            domain='PloneMeeting',
            context=self.request)
        errors = invariants.validate(data)
        self.request.set('validate_attendees_done', False)
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].message, secret_error_msg)
        self.request.set('validate_attendees_done', False)
        meeting_voters.insert(0, voter0)
        self.assertEqual(invariants.validate(data), ())

    def test_pm_CanNotSetAbsentAnAttendeeThatVoted(self):
        """ """
        self.changeUser('pmManager')
        meeting, public_item, secret_item = \
            self._createMeetingWithVotes(include_yes=False)

        # byebye person on public_item
        person1 = self.portal.contacts.get('person1')
        hp1 = person1.get_held_positions()[0]
        hp1_uid = hp1.UID()
        byebye_form = public_item.restrictedTraverse('@@item_byebye_attendee_form')
        byebye_form.meeting = meeting
        byebye_form.person_uid = hp1_uid
        byebye_form.not_present_type = 'excused'
        byebye_form.apply_until_item_number = '200'
        self.assertEqual(byebye_form._doApply(),
                         u'Can not set "Absent (excused)" a person that voted on an item!')
        # was not set excused
        self.assertFalse(meeting.get_item_excused(by_persons=True))

        # now remove hp1 from public_item and secret_item
        public_votes = public_item.get_item_votes()[0]
        public_votes['voters'][hp1_uid] = NOT_ENCODED_VOTE_VALUE
        meeting.set_item_public_vote(public_item, public_votes, 0)
        transaction.commit()
        # not done because could not be done on secret_item
        self.assertEqual(
            byebye_form._doApply(),
            u'Can not set "Absent (excused)" a person that voted on an item!\n'
            u'Please check item number 2 at http://nohost/plone/Members/pmManager'
            u'/mymeetings/%s/o3.' % self.meetingConfig.getId())
        self.assertFalse(meeting.get_item_excused(by_persons=True))
        # encode secret votes
        secret_votes = secret_item.get_item_votes()[0]
        secret_votes['votes']['yes'] = 0
        meeting.set_item_secret_vote(secret_item, secret_votes, 0)
        transaction.commit()
        self.assertIsNone(byebye_form._doApply())
        self.assertEqual(
            sorted(meeting.get_item_excused(by_persons=True)[hp1_uid]),
            sorted([public_item.UID(), secret_item.UID()]))

    def test_pm_EncodePublicVotesForm(self):
        """ """
        cfg = self.meetingConfig
        self.changeUser('pmManager')
        meeting, public_item, secret_item = \
            self._createMeetingWithVotes(include_yes=False)

        # encode votes form
        person1 = self.portal.contacts.get('person1')
        hp1 = person1.get_held_positions()[0]
        hp1_uid = hp1.UID()
        person2 = self.portal.contacts.get('person2')
        hp2 = person2.get_held_positions()[0]
        hp2_uid = hp2.UID()
        person3 = self.portal.contacts.get('person3')
        hp3 = person3.get_held_positions()[0]
        hp3_uid = hp3.UID()
        person4 = self.portal.contacts.get('person4')
        hp4 = person4.get_held_positions()[0]
        hp4_uid = hp4.UID()
        self.request['PUBLISHED'] = public_item
        votes_form = public_item.restrictedTraverse('@@item-encode-votes-form')
        votes_form.meeting = meeting
        # change vote to all 'no'
        votes_form.votes = [{'voter_uid': hp1_uid, 'vote_value': 'no'},
                            {'voter_uid': hp2_uid, 'vote_value': 'no'},
                            {'voter_uid': hp3_uid, 'vote_value': 'no'},
                            {'voter_uid': hp4_uid, 'vote_value': 'no'}]
        votes_form.vote_number = 0
        self.request.form['vote_number'] = 0
        votes_form.label = u"My label"
        votes_form.linked_to_previous = False
        votes_form.apply_until_item_number = u'0'
        votes_form.update()
        # only for MeetingManagers
        self.changeUser('pmCreator1')
        self.assertRaises(Unauthorized, votes_form._doApply)
        self.changeUser('pmManager')
        self.assertEqual(public_item.get_vote_count(meeting, 'yes'), 2)
        votes_form.update()
        votes_form._doApply()
        transaction.commit()
        # votes were updated
        self.assertEqual(public_item.get_vote_count(meeting, 'yes'), 0)
        self.assertEqual(public_item.get_vote_count(meeting, 'no'), 4)
        # disable voter, it is still displayed when editing voters for item
        # deactivated, form still OK
        votes_form._finished = False
        self.changeUser('siteadmin')
        self.do(hp4, 'deactivate')
        cleanRamCache()
        self.changeUser('pmManager')
        self.assertTrue(hp4_uid in votes_form.render())
        # remove usages from hp
        hp4.usages = []
        cleanRamCache()
        self.assertTrue(hp4_uid in votes_form.render())
        # make hp no more selectable
        ordered_contacts = list(cfg.getOrderedContacts())
        ordered_contacts.remove(hp4_uid)
        cfg.setOrderedContacts(ordered_contacts)
        cleanRamCache()
        self.assertTrue(hp4_uid in votes_form.render())

    def test_pm_EncodePublicVotesFormLinkedToPrevious(self):
        """ """
        self.changeUser('pmManager')
        meeting, public_item, secret_item = \
            self._createMeetingWithVotes(include_yes=False)

        # encode votes form
        person1 = self.portal.contacts.get('person1')
        hp1 = person1.get_held_positions()[0]
        hp1_uid = hp1.UID()
        person2 = self.portal.contacts.get('person2')
        hp2 = person2.get_held_positions()[0]
        hp2_uid = hp2.UID()
        person3 = self.portal.contacts.get('person3')
        hp3 = person3.get_held_positions()[0]
        hp3_uid = hp3.UID()
        person4 = self.portal.contacts.get('person4')
        hp4 = person4.get_held_positions()[0]
        hp4_uid = hp4.UID()
        votes_form = public_item.restrictedTraverse('@@item-encode-votes-form')
        votes_form.meeting = meeting
        # there are 'yes' votes so not able to link to previous
        self.assertEqual(public_item.get_vote_count(meeting, 'yes'), 2)
        load_view = public_item.restrictedTraverse('@@load_item_assembly_and_signatures')
        load_view._update()
        self.assertFalse(load_view.show_add_vote_linked_to_previous_icon(vote_number=0))

        # make linked vote addable
        votes_form.votes = [{'voter_uid': hp1_uid, 'vote_value': 'no'},
                            {'voter_uid': hp2_uid, 'vote_value': 'abstain'},
                            {'voter_uid': hp3_uid, 'vote_value': NOT_ENCODED_VOTE_VALUE},
                            {'voter_uid': hp4_uid, 'vote_value': NOT_ENCODED_VOTE_VALUE}]
        votes_form.vote_number = 0
        votes_form.label = u"My label"
        votes_form.linked_to_previous = False
        votes_form.apply_until_item_number = u'0'
        votes_form._doApply()
        transaction.commit()
        load_view._update()
        self.assertTrue(load_view.show_add_vote_linked_to_previous_icon(vote_number=0))

        # add linked vote
        self.request.set('form.widgets.linked_to_previous', True)
        self.request.set('vote_number', 1)
        # votes default only show encodable values for hp3/hp4
        self.assertEqual(
            votes_default(public_item),
            [{'vote_value': NOT_ENCODED_VOTE_VALUE,
              'voter': hp3_uid,
              'voter_uid': hp3_uid},
             {'vote_value': NOT_ENCODED_VOTE_VALUE,
              'voter': hp4_uid,
              'voter_uid': hp4_uid}])
        # apply linked vote
        votes_form.vote_number = 1
        votes_form.label = u"My label 1"
        votes_form.linked_to_previous = True
        votes_form.apply_until_item_number = u'0'
        votes_form.votes = [{'voter_uid': hp3_uid, 'vote_value': 'yes'},
                            {'voter_uid': hp4_uid, 'vote_value': NOT_ENCODED_VOTE_VALUE}]
        votes_form._doApply()
        # 2 encoded votes
        item_votes = public_item.get_item_votes()
        self.assertEqual(len(item_votes), 2)
        # votes not useable in vote_number 0 or 1 are marked NOT_VOTABLE_LINKED_TO_VALUE
        self.assertEqual(item_votes[0]['voters'][hp3_uid], NOT_VOTABLE_LINKED_TO_VALUE)
        self.assertEqual(item_votes[1]['voters'][hp1_uid], NOT_VOTABLE_LINKED_TO_VALUE)
        self.assertEqual(item_votes[1]['voters'][hp2_uid], NOT_VOTABLE_LINKED_TO_VALUE)
        # if not encoded in vote_number 0 and 1, some values appear in both
        self.assertEqual(item_votes[0]['voters'][hp4_uid], NOT_ENCODED_VOTE_VALUE)
        self.assertEqual(item_votes[1]['voters'][hp4_uid], NOT_ENCODED_VOTE_VALUE)
        # finally encode hp4_uid
        votes_form.votes = [{'voter_uid': hp3_uid, 'vote_value': 'yes'},
                            {'voter_uid': hp4_uid, 'vote_value': 'yes'}]
        votes_form._doApply()
        transaction.commit()
        item_votes = public_item.get_item_votes()
        self.assertEqual(item_votes[0]['voters'][hp1_uid], 'no')
        self.assertEqual(item_votes[0]['voters'][hp2_uid], 'abstain')
        self.assertEqual(item_votes[0]['voters'][hp3_uid], NOT_VOTABLE_LINKED_TO_VALUE)
        self.assertEqual(item_votes[0]['voters'][hp4_uid], NOT_VOTABLE_LINKED_TO_VALUE)
        self.assertEqual(item_votes[1]['voters'][hp1_uid], NOT_VOTABLE_LINKED_TO_VALUE)
        self.assertEqual(item_votes[1]['voters'][hp2_uid], NOT_VOTABLE_LINKED_TO_VALUE)
        self.assertEqual(item_votes[1]['voters'][hp3_uid], 'yes')
        self.assertEqual(item_votes[1]['voters'][hp4_uid], 'yes')

    def _setUpSecretVoteForm(self, secret_item, meeting):
        """ """
        votes_form = secret_item.restrictedTraverse(
            '@@item-encode-secret-votes-form')
        self.request['PUBLISHED'] = secret_item
        votes_form.meeting = meeting
        votes_form.votes = [
            {'vote_value': 'yes', 'vote_count': 0, 'vote_value_id': 'yes'},
            {'vote_value': 'no', 'vote_count': 4, 'vote_value_id': 'no'},
            {'vote_value': 'abstain', 'vote_count': 0, 'vote_value_id': 'abstain'}]
        self.request.form['vote_number'] = 0
        votes_form.vote_number = 0
        votes_form.label = u"My label"
        votes_form.linked_to_previous = False
        votes_form.apply_until_item_number = u'0'
        return votes_form

    def _setUpPublicVoteForm(self, public_item, meeting):
        """ """
        person1 = self.portal.contacts.get('person1')
        hp1 = person1.get_held_positions()[0]
        hp1_uid = hp1.UID()
        person2 = self.portal.contacts.get('person2')
        hp2 = person2.get_held_positions()[0]
        hp2_uid = hp2.UID()
        person3 = self.portal.contacts.get('person3')
        hp3 = person3.get_held_positions()[0]
        hp3_uid = hp3.UID()
        person4 = self.portal.contacts.get('person4')
        hp4 = person4.get_held_positions()[0]
        hp4_uid = hp4.UID()
        votes_form = public_item.restrictedTraverse(
            '@@item-encode-votes-form')
        votes_form.meeting = meeting
        # change vote to all 'no'
        votes_form.votes = [{'voter_uid': hp1_uid, 'vote_value': 'no'},
                            {'voter_uid': hp2_uid, 'vote_value': 'no'},
                            {'voter_uid': hp3_uid, 'vote_value': 'no'},
                            {'voter_uid': hp4_uid, 'vote_value': 'no'}]
        votes_form.vote_number = 0
        self.request.form['vote_number'] = 0
        votes_form.label = u"My label"
        votes_form.linked_to_previous = False
        votes_form.apply_until_item_number = u'0'
        return votes_form

    def test_pm_EncodeSecretVotesForm(self):
        """ """
        self.changeUser('pmManager')
        meeting, public_item, secret_item = \
            self._createMeetingWithVotes(include_yes=False)

        # encode votes form
        votes_form = self._setUpSecretVoteForm(secret_item, meeting)
        # only for MeetingManagers
        self.changeUser('pmCreator1')
        self.assertRaises(Unauthorized, votes_form._doApply)
        self.changeUser('pmManager')
        self.assertEqual(secret_item.get_vote_count(meeting, 'yes'), 1)
        votes_form.update()
        votes_form._doApply()
        transaction.commit()
        # votes were updated
        self.assertEqual(secret_item.get_vote_count(meeting, 'yes'), 0)
        self.assertEqual(secret_item.get_vote_count(meeting, 'no'), 4)

    def test_pm_EncodeSecretVotesFormLinkedToPrevious(self):
        """ """
        self.changeUser('pmManager')
        meeting, public_item, secret_item = \
            self._createMeetingWithVotes(include_yes=False)

        # there are 'yes' votes so not able to link to previous
        self.assertEqual(secret_item.get_vote_count(meeting, 'yes'), 1)
        load_view = secret_item.restrictedTraverse('@@load_item_assembly_and_signatures')
        load_view._update()
        self.assertFalse(load_view.show_add_vote_linked_to_previous_icon(vote_number=0))

        # make linked vote addable
        votes_form = secret_item.restrictedTraverse('@@item-encode-secret-votes-form')
        votes_form.meeting = meeting
        votes_form.votes = [
            {'vote_value': 'yes', 'vote_count': 0, 'vote_value_id': 'yes'},
            {'vote_value': 'no', 'vote_count': 2, 'vote_value_id': 'no'},
            {'vote_value': 'abstain', 'vote_count': 0, 'vote_value_id': 'abstain'}]
        votes_form.vote_number = 0
        votes_form.label = u"My label"
        votes_form.linked_to_previous = False
        votes_form.apply_until_item_number = u'0'
        # item modified when applied
        modified = secret_item.modified()
        votes_form._doApply()
        self.assertNotEqual(modified, secret_item.modified())
        load_view._update()
        transaction.commit()
        load_view._update()
        self.assertTrue(load_view.show_add_vote_linked_to_previous_icon(vote_number=0))

        # add linked vote
        self.request.set('form.widgets.linked_to_previous', True)
        self.request.set('vote_number', 1)
        # votes default only show encodable values for hp3/hp4
        self.assertEqual(
            secret_votes_default(secret_item),
            [{'vote_value': 'yes', 'vote_count': 0, 'vote_value_id': 'yes'}])
        # apply linked vote
        votes_form.vote_number = 1
        votes_form.label = u"My label 1"
        votes_form.linked_to_previous = True
        votes_form.apply_until_item_number = u'0'
        votes_form.votes = [{'vote_value': 'yes', 'vote_count': 1, 'vote_value_id': 'yes'}]
        votes_form._doApply()
        # 2 encoded votes
        item_votes = secret_item.get_item_votes()
        self.assertEqual(len(item_votes), 2)

    def test_pm_EncodeSecretVotesFormInvariant(self):
        """The validate_votes invariant check that encoded values do not
           overflow maximum number of votes."""

        class DummyData(object):
            def __init__(self, context, votes, vote_number=0):
                self.__context__ = context
                self.votes = votes
                self.vote_number = vote_number

        self.changeUser('pmManager')
        meeting, public_item, secret_item = \
            self._createMeetingWithVotes(include_yes=False)

        # one vote, maximum voter is 4
        invariant = IEncodeSecretVotes.getTaggedValue('invariants')[0]
        votes = [
            {'vote_value': 'yes', 'vote_count': 0, 'vote_value_id': 'yes'},
            {'vote_value': 'no', 'vote_count': 2, 'vote_value_id': 'no'},
            {'vote_value': 'abstain', 'vote_count': 0, 'vote_value_id': 'abstain'}]
        data = DummyData(secret_item, votes)
        self.assertIsNone(invariant(data))
        # validation fails if total > 4
        error_msg = translate('error_can_not_encode_more_than_max_voters',
                              mapping={'max_voters': 4},
                              domain='PloneMeeting',
                              context=self.request)
        votes = [
            {'vote_value': 'yes', 'vote_count': 2, 'vote_value_id': 'yes'},
            {'vote_value': 'no', 'vote_count': 2, 'vote_value_id': 'no'},
            {'vote_value': 'abstain', 'vote_count': 2, 'vote_value_id': 'abstain'}]
        data = DummyData(secret_item, votes)
        with self.assertRaises(Invalid) as cm:
            invariant(data)
        self.assertEqual(cm.exception.message, error_msg)

        # linked vote
        self.request.form['form.widgets.linked_to_previous'] = True
        # already 4 votes, encoding 0 pass
        votes = [{'vote_value': 'yes', 'vote_count': 0, 'vote_value_id': 'yes'}, ]
        data = DummyData(secret_item, votes, vote_number=1)
        self.assertIsNone(invariant(data))
        # already 4 votes, encoding 1 would do 5 and fails
        votes = [{'vote_value': 'yes', 'vote_count': 1, 'vote_value_id': 'yes'}, ]
        data = DummyData(secret_item, votes, vote_number=1)
        with self.assertRaises(Invalid) as cm:
            invariant(data)
        self.assertEqual(cm.exception.message, error_msg)

        # when used in an overlay, the PMNumberWidget number brower validation
        # is not correctly done, we could get values other than integers...
        error_msg = translate('error_some_values_are_not_integers',
                              domain='PloneMeeting',
                              context=self.request)
        votes = [
            {'vote_value': 'yes', 'vote_count': None, 'vote_value_id': 'yes'},
            {'vote_value': 'no', 'vote_count': 2, 'vote_value_id': 'no'},
            {'vote_value': 'abstain', 'vote_count': 2, 'vote_value_id': 'abstain'}]
        data = DummyData(secret_item, votes)
        with self.assertRaises(Invalid) as cm:
            invariant(data)
        self.assertEqual(cm.exception.message, error_msg)

    def test_pm_ItemVotesWhenItemRemovedFromMeeting(self):
        """Ensure Meeting.item_votes correctly wiped out when item removed from meeting."""
        self.changeUser('pmManager')
        meeting, public_item, yes_public_item, secret_item, yes_secret_item = \
            self._createMeetingWithVotes()

        public_item_uid = public_item.UID()
        self.assertTrue(public_item_uid in meeting.item_votes)
        secret_item_uid = secret_item.UID()
        self.assertTrue(secret_item_uid in meeting.item_votes)
        self.backToState(public_item, 'validated')
        self.assertFalse(public_item_uid in meeting.item_votes)
        self.backToState(secret_item, 'validated')
        self.assertFalse(secret_item_uid in meeting.item_votes)

    def test_pm_DisplayMeetingItemVoters(self):
        """The view that displays items for which votes are (not) completed."""
        self.changeUser('pmManager')
        meeting, public_item, yes_public_item, secret_item, yes_secret_item = \
            self._createMeetingWithVotes()

        # for now every voters voted
        # non voted
        non_voted_view = meeting.restrictedTraverse('@@display-meeting-item-voters')
        self.assertEqual(non_voted_view.get_non_voted_items(),
                         {'no_vote': [], 'secret': [], 'public': []})
        non_voted_view()
        # voted
        voted_view = meeting.restrictedTraverse('@@display-meeting-item-voters')
        voted_view.show_voted_items = True
        voted_items = voted_view.get_voted_items()
        self.assertTrue(public_item in voted_items['public'])
        self.assertTrue(yes_public_item in voted_items['public'])
        self.assertTrue(secret_item in voted_items['secret'])
        self.assertTrue(yes_secret_item in voted_items['secret'])
        voted_view()

        # remove one voter on public_item and secret_item
        # public vote
        voters = meeting.get_voters()
        public_votes = public_item.get_item_votes()[0]
        public_votes['voters'][voters[0]] = NOT_ENCODED_VOTE_VALUE
        meeting.set_item_public_vote(public_item, public_votes, 0)
        transaction.commit()
        # secret vote
        secret_votes = secret_item.get_item_votes()[0]
        secret_votes['votes']['yes'] = 0
        meeting.set_item_secret_vote(secret_item, secret_votes, 0)
        transaction.commit()
        # non voted
        non_voted_items = non_voted_view.get_non_voted_items()
        self.assertTrue(public_item in non_voted_items['public'])
        self.assertTrue(secret_item in non_voted_items['secret'])
        non_voted_view()
        # voted
        voted_items = voted_view.get_voted_items()
        self.assertFalse(public_item in voted_items['public'])
        self.assertFalse(secret_item in voted_items['secret'])
        voted_view()
        # no_vote
        public_item.setPollType('no_vote')
        del meeting.item_votes[public_item.UID()]
        transaction.commit()
        self.assertEqual(public_item.get_item_votes(), [])
        self.assertTrue(public_item in voted_view.get_non_voted_items()['no_vote'])
        self.assertTrue(public_item in voted_view.get_voted_items()['no_vote'])
        non_voted_view()
        voted_view()

    def test_pm_ChangePollTypeView(self):
        """The view that let's change MeetingItem.pollType on item view.
           It manage changes, but also avoid to change from a "secret" mode
           to a "public" mode if some votes are already encoded.
           The validation also work from MeetingItem edit form."""
        self._enableField('pollType')
        self.changeUser('pmManager')
        meeting, public_item, yes_public_item, secret_item, yes_secret_item = \
            self._createMeetingWithVotes()
        pt_view = secret_item.restrictedTraverse('@@item-polltype')
        self.assertTrue(pt_view.selectablePollTypes())
        change_pt_view = secret_item.restrictedTraverse('@@change-item-polltype')
        # try to change to an unexisting value
        self.assertRaises(KeyError, change_pt_view, "unexisting")
        self.assertRaises(KeyError, secret_item.validate_pollType, "unexisting")
        # can not switch to no_vote if votes encoded
        self.assertTrue(secret_item.get_item_votes())
        original_poll_type = secret_item.getPollType()
        self.assertEqual(original_poll_type, 'secret')
        change_pt_view("no_vote")
        self.assertTrue(secret_item.validate_pollType("no_vote"))
        self.assertEqual(secret_item.getPollType(), original_poll_type)
        # can not switch to a "public" mode vote
        change_pt_view("freehand")
        self.assertEqual(secret_item.getPollType(), original_poll_type)
        self.assertTrue(secret_item.validate_pollType("freehand"))
        # but can change to a vote is same mode, "secret"
        self.failIf(secret_item.validate_pollType("secret_separated"))
        change_pt_view("secret_separated")
        self.assertNotEqual(secret_item.getPollType(), original_poll_type)
        self.assertEqual(secret_item.getPollType(), "secret_separated")

    def test_pm_AsyncLoadMeetingAssemblyAndSignatures(self):
        """The @@load_meeting_assembly_and_signatures will load attendees
           details on the meeting view."""
        self.changeUser('pmManager')
        meeting, public_item, yes_public_item, secret_item, yes_secret_item = \
            self._createMeetingWithVotes()
        view = meeting.restrictedTraverse('@@load_meeting_assembly_and_signatures')
        rendered = view()
        # MeetingManager see the attendees
        self.assertTrue("Monsieur Person1FirstName Person1LastName, Assembly member 1" in rendered)
        # and voters actions
        self.assertTrue("@@display-meeting-item-voters" in rendered)
        # only attendees for users, not voters actions
        self.changeUser('pmCreator1')
        rendered = view()
        self.assertTrue("Monsieur Person1FirstName Person1LastName, Assembly member 1" in rendered)
        self.assertFalse("@@display-meeting-item-voters" in rendered)

        # passing a cache_date will invalidate cache
        self.assertFalse('observations' in view.used_attrs)
        self._enableField('assembly_observations', related_to='Meeting')
        # not changed for now as cachekey still valid
        self.assertFalse('assembly_observations' in view())
        # managed by JS
        self.request['cache_date'] = "Tue Nov 08 2022 14:41:49 GMT+0100"
        self.assertTrue('assembly_observations' in view())
        # still correct when cache_date no more used
        self.request['cache_date'] = None
        self.assertTrue('assembly_observations' in view())

    def test_pm_AsyncLoadItemAssemblyAndSignatures(self):
        """The @@load_item_assembly_and_signatures will load attendees
           details on the item view."""
        self.changeUser('pmManager')
        self.create('Meeting')
        item = self.create('MeetingItem')
        self.presentItem(item)
        view = item.restrictedTraverse('@@load_item_assembly_and_signatures')
        rendered = view()
        # MeetingManager see and may manage
        self.assertTrue("Monsieur Person1FirstName Person1LastName, Assembly member 1" in rendered)
        manage_vote_action = "item-encode-votes-form?vote_number:int=0"
        manage_attendee_action = "item_byebye_attendee_form?person_uid="
        manage_signatory_action = "item_redefine_signatory_form?person_uid="
        self.assertTrue(manage_vote_action in rendered)
        self.assertTrue(manage_attendee_action in rendered)
        self.assertTrue(manage_signatory_action in rendered)

        # other users may see but not manage
        self.changeUser('pmCreator1')
        view = item.restrictedTraverse('@@load_item_assembly_and_signatures')
        rendered = view()
        self.assertTrue("Monsieur Person1FirstName Person1LastName, Assembly member 1" in rendered)
        self.assertFalse(manage_vote_action in rendered)
        self.assertFalse(manage_attendee_action in rendered)
        self.assertFalse(manage_signatory_action in rendered)

        # passing a cache_date will invalidate cache
        self.assertFalse('assembly_guests' in view.used_meeting_attrs)
        self._enableField('assembly_guests', related_to='Meeting')
        # not changed for now as cachekey still valid
        self.assertFalse('Assembly guests' in view())
        # managed by JS
        self.request['cache_date'] = "Tue Nov 08 2022 14:41:49 GMT+0100"
        self.assertTrue('Assembly guests' in view())
        # still correct when cache_date no more used
        self.request['cache_date'] = None
        self.assertTrue('Assembly guests' in view())

    def test_pm_EncodeVotesForSeveralItems(self):
        """Votes may be encoded for several items but only relevant items will be updated."""
        self.changeUser('pmManager')
        meeting, public_item, yes_public_item, secret_item, yes_secret_item, no_vote_item = \
            self._createMeetingWithVotes(include_no_vote=True)
        # for now public_item and yes_public_item votes are different
        self.assertNotEqual(public_item.get_item_votes(), yes_public_item.get_item_votes())
        secret_item_votes = secret_item.get_item_votes()
        yes_secret_item_votes = yes_secret_item.get_item_votes()
        self.request['PUBLISHED'] = public_item
        votes_form = public_item.restrictedTraverse('@@item-encode-votes-form')
        self.request.form['vote_number'] = 0
        votes_form.meeting = public_item.getMeeting()
        votes_form.update()
        votes_form.votes = votes_form.widgets['votes'].value
        votes_form.linked_to_previous = False
        votes_form.vote_number = 0
        votes_form.apply_until_item_number = u'500'
        # item modified when applied
        modified = public_item.modified()
        votes_form._doApply()
        transaction.commit()
        self.assertNotEqual(modified, public_item.modified())
        # votes were updated for yes_public_item but not for secret items
        self.assertEqual(public_item.get_item_votes(), yes_public_item.get_item_votes())
        self.assertEqual(secret_item_votes, secret_item.get_item_votes())
        self.assertEqual(yes_secret_item_votes, yes_secret_item.get_item_votes())
        self.assertEqual(no_vote_item.get_item_votes(), [])
        self.assertEqual(no_vote_item.get_item_votes(vote_number=0), [])
        # relevant messages are displayed to the user
        messages = IStatusMessage(self.request).show()
        self.assertEqual(messages[-2].message, u'Votes have been encoded for items "1 & 2".')
        self.assertEqual(messages[-1].message, u'Votes could not be updated for items "3, 4 & 5".')

    def test_pm_EncodeVotesByVotingGroup(self):
        """Just check that voting_group functionnality is working."""
        # create an organization outside own_org
        self.changeUser('siteadmin')
        outside_org = api.content.create(
            container=self.portal.contacts,
            type='organization',
            id='org-outside-own-org',
            title='Organization outside own org')
        # define voting_group for some voters
        person1 = self.portal.contacts.get('person1')
        hp1 = person1.get_held_positions()[0]
        hp1.voting_group = self._relation(outside_org)

        # display form and check that elements are there
        self.changeUser('pmManager')
        meeting, public_item, yes_public_item, secret_item, yes_secret_item = \
            self._createMeetingWithVotes()
        self.request['PUBLISHED'] = public_item
        votes_form = public_item.restrictedTraverse('@@item-encode-votes-form')
        self.request.form['vote_number'] = 0
        votes_form.meeting = public_item.getMeeting()
        votes_form.update()
        self.request.form['ajax_load'] = 'dummy'
        rendered_form = votes_form.render()
        # select values are there
        self.assertTrue("<span>All</span>:" in rendered_form)
        self.assertTrue("<span>Organization outside own org</span>:" in rendered_form)
        self.assertTrue("<span>Others</span>:" in rendered_form)
        self.assertTrue(
            '<tr class="datagridwidget-row required org-outside-own-org row-1" data-index="0">'
            in rendered_form)
        # when no voting_group defined for any voter, controls are not there
        hp1.voting_group = None
        # clear cache
        self.request['_build_voting_groups'] = None
        votes_form.update()
        rendered_form = votes_form.render()
        # select values are there
        self.assertTrue("<span>All</span>:" in rendered_form)
        self.assertFalse("<span>Organization outside own org</span>:" in rendered_form)
        self.assertFalse("<span>Others</span>:" in rendered_form)
        self.assertFalse(
            '<tr class="datagridwidget-row required org-outside-own-org row-1" data-index="0">'
            in rendered_form)

    def test_pm_ChangeSecretVotePollType(self):
        """Change poll_type on a secret vote."""
        self.changeUser('pmManager')
        meeting, public_item, yes_public_item, secret_item, yes_secret_item = \
            self._createMeetingWithVotes()
        # not changed when encoded
        self.assertTrue(secret_item.get_vote_is_secret(meeting, 0))
        view = secret_item.restrictedTraverse('@@change-item-vote-polltype')
        view(0, 'freehand')
        self.assertTrue(secret_item.get_vote_is_secret(meeting, 0))
        # encode votes form
        self._setUpSecretVoteForm(secret_item, meeting)
        # "apply_until_item_number" field is useable
        self.assertFalse(_should_disable_apply_until_item_number(secret_item))

        # the view on item is rendering
        votes_view = secret_item.restrictedTraverse('@@load_item_assembly_and_signatures')
        rendered = votes_view()
        self.assertTrue("secret-vote" in rendered)
        self.assertFalse("public-vote" in rendered)

        # add an empty vote and change it's poll_type
        secret_vote = secret_item.get_item_votes(0)
        secret_vote['votes']['yes'] = 0
        secret_vote['votes']['abstain'] = 0
        secret_vote['votes']['no'] = 0
        meeting.set_item_secret_vote(secret_item, secret_vote, 1)
        transaction.commit()
        view(1, 'freehand')
        # commit so meeting.item_votes._p_mtime is updated
        transaction.commit()
        # "apply_until_item_number" field would be disabled
        self.assertTrue(_should_disable_apply_until_item_number(secret_item))

        # vote_number 1 is now public
        self.assertTrue(secret_item.get_vote_is_secret(meeting, 0))
        self.assertFalse(secret_item.get_vote_is_secret(meeting, 1))
        self.assertFalse("voters" in secret_item.get_item_votes(0))
        self.assertTrue("voters" in secret_item.get_item_votes(1))
        # the view on item is rendering
        rendered = votes_view()
        self.assertTrue("secret-vote" in rendered)
        self.assertTrue("public-vote" in rendered)

    def test_pm_ChangePublicVotePollType(self):
        """Change poll_type on a public vote."""
        self.changeUser('pmManager')
        meeting, public_item, yes_public_item, secret_item, yes_secret_item = \
            self._createMeetingWithVotes()
        # not changed when encoded
        self.assertFalse(public_item.get_vote_is_secret(meeting, 0))
        view = public_item.restrictedTraverse('@@change-item-vote-polltype')
        view(0, 'secret')
        self.assertFalse(public_item.get_vote_is_secret(meeting, 0))
        # encode votes form
        self._setUpPublicVoteForm(public_item, meeting)
        # "apply_until_item_number" field is useable
        self.assertFalse(_should_disable_apply_until_item_number(public_item))

        # the view on item is rendering
        votes_view = public_item.restrictedTraverse(
            '@@load_item_assembly_and_signatures')
        rendered = votes_view()
        self.assertTrue("public-vote" in rendered)
        self.assertFalse("secret-vote" in rendered)

        # add an empty vote and change it's poll_type
        public_vote = public_item.get_item_votes(0)
        public_vote["voters"] = {hp_uid: NOT_ENCODED_VOTE_VALUE
                                 for hp_uid in public_vote["voters"]}
        meeting.set_item_public_vote(public_item, public_vote, 1)
        transaction.commit()
        view(1, 'secret')
        # commit so meeting.item_votes._p_mtime is updated
        transaction.commit()
        # "apply_until_item_number" field would be disabled
        self.assertTrue(_should_disable_apply_until_item_number(public_item))

        # vote_number 1 is now secret
        self.assertFalse(public_item.get_vote_is_secret(meeting, 0))
        self.assertTrue(public_item.get_vote_is_secret(meeting, 1))
        self.assertTrue("voters" in public_item.get_item_votes(0))
        self.assertFalse("voters" in public_item.get_item_votes(1))
        # the view on item is rendering
        votes_view = public_item.restrictedTraverse(
            '@@load_item_assembly_and_signatures')
        rendered = votes_view()
        self.assertTrue("public-vote" in rendered)
        self.assertTrue("secret-vote" in rendered)

    def test_pm_ItemGetVotesResult(self):
        """Field MeetingItem.votesResult accessor is overrided to handle
           votes result generated text."""
        cfg = self.meetingConfig
        # disable itempublished so we play nice with plugins
        if self._check_wfa_available(['no_publication']):
            self._activate_wfas(('no_publication', ), keep_existing=False)
        self._removeConfigObjectsFor(cfg)
        self._enableField('votesResult')
        self._enableField('votesResult_after_motivation')
        self.changeUser('pmCreator1')
        item = self.create('MeetingItem')
        self.assertFalse(item.mayQuickEdit('votesResult'))
        self.assertTrue(may_view_field(item, 'votesResult'))
        self.changeUser('pmManager')
        self.assertFalse(item.mayQuickEdit('votesResult'))
        self.assertTrue(may_view_field(item, 'votesResult'))

        # get outside meeting
        self.assertEqual(cfg.getVotesResultTALExpr(), '')
        self.assertEqual(item.getVotesResult(), '')
        self.assertEqual(item.getVotesResult(real=True), '')
        self.assertFalse(isinstance(item.getVotesResult(), unicode))
        cfg.setVotesResultTALExpr(
            'python: pm_utils.print_votes(item, include_total_voters=True)')
        cleanRamCache()
        # not computed when not in a meeting
        self.assertEqual(item.getVotesResult(), '')
        self.assertFalse(isinstance(item.getVotesResult(), unicode))
        self.assertEqual(item.getVotesResult(real=True), '')

        # get in meeting
        meeting, public_item, yes_public_item, secret_item, yes_secret_item = \
            self._createMeetingWithVotes()
        # votes are editable as soon as in a meeting (presented)
        self.assertEqual(public_item.query_state(), 'presented')
        self.assertTrue(public_item.mayQuickEdit('votesResult'))
        self.assertTrue(may_view_field(public_item, 'votesResult'))
        self.assertEqual(
            public_item.getVotesResult(),
            '<p>Il y a 4 votants.</p><p>Par 2 voix pour, une voix contre '
            'et une abstention,</p>')
        self.assertEqual(
            yes_public_item.getVotesResult(),
            "<p>Il y a 4 votants.</p><p>\xc3\x80 l'unanimit\xc3\xa9,</p>")
        self.assertEqual(
            secret_item.getVotesResult(),
            '<p>Il y a 4 votants.</p><p>Au scrutin secret,</p>'
            '<p>Par une voix pour, une voix contre et 2 abstentions,</p>')
        self.assertEqual(
            yes_secret_item.getVotesResult(),
            "<p>Il y a 4 votants.</p><p>Au scrutin secret,</p>"
            "<p>\xc3\x80 l'unanimit\xc3\xa9,</p>")

        # freeze the meeting and set values
        self.freezeMeeting(meeting)
        self.assertEqual(public_item.query_state(), 'itemfrozen')
        self.assertTrue(public_item.mayQuickEdit('votesResult'))
        self.assertTrue(secret_item.mayQuickEdit('votesResult'))

        # when a value is set, then it is used
        self.assertFalse(public_item.getVotesResult(real=True))
        self.assertFalse(secret_item.getVotesResult(real=True))
        public_item.setVotesResult('<p>Custom public text.</p>')
        self.assertEqual(public_item.getVotesResult(), '<p>Custom public text.</p>')
        self.assertEqual(public_item.getVotesResult(real=True), '<p>Custom public text.</p>')
        secret_item.setVotesResult('<p>Custom secret text.</p>')
        self.assertEqual(secret_item.getVotesResult(), '<p>Custom secret text.</p>')
        self.assertEqual(secret_item.getVotesResult(real=True), '<p>Custom secret text.</p>')

        # decide item, still editable until meeting is closed
        self.decideMeeting(meeting)
        self.assertEqual(public_item.query_state(), 'itemfrozen')
        self.assertTrue(public_item.mayQuickEdit('votesResult'))
        self.assertTrue(secret_item.mayQuickEdit('votesResult'))
        self.do(public_item, 'accept')
        self.assertEqual(public_item.query_state(), 'accepted')
        self.assertTrue(public_item.mayQuickEdit('votesResult'))
        self.closeMeeting(meeting)
        self.assertEqual(public_item.query_state(), 'accepted')
        self.assertFalse(public_item.mayQuickEdit('votesResult'))

        # wrong expression will not break the view, if result is not html
        # a portal_messag is displayed
        IStatusMessage(self.request).show()
        item.setVotesResult('')
        public_item.setVotesResult('')
        cfg.setVotesResultTALExpr("string:not html")
        cleanRamCache()
        # no message as item not in a meeting
        item.restrictedTraverse('base_view')()
        self.assertFalse(IStatusMessage(self.request).show())
        # message as public_item in a meeting
        public_item.restrictedTraverse('base_view')()
        self.assertEqual(IStatusMessage(self.request).show()[0].message,
                         u'Votes result is not HTML')
        # is_all_count is also available on pm_utils
        cfg.setVotesResultTALExpr(
            "python: pm_utils.is_all_count(item) and '<p>All OK</p>' or '<p>Not all OK</p>'")
        cleanRamCache()
        self.assertEqual(public_item.getVotesResult(), '<p>Not all OK</p>')
        self.assertEqual(yes_secret_item.getVotesResult(), '<p>All OK</p>')


def test_suite():
    from unittest import makeSuite
    from unittest import TestSuite
    suite = TestSuite()
    suite.addTest(makeSuite(testVotes, prefix='test_pm_'))
    return suite
